# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: © 2019 Tobias Gruetzmacher
# SPDX-FileCopyrightText: © 2019 Daniel Ring
import json

from .. import scraper, util


class MangaDex(scraper.ParserScraper):
    multipleImagesPerStrip = True

    def __init__(self, name, mangaId):
        super().__init__('MangaDex/' + name)

        baseUrl = 'https://api.mangadex.org/'
        self.url = baseUrl + 'manga/%s' % mangaId
        self.chaptersUrl = baseUrl + 'manga/%s/feed?translatedLanguage[]=en&order[chapter]=desc&limit=500' % mangaId
        self.stripUrl = baseUrl + 'chapter/%s'
        self.cdnUrl = baseUrl + 'at-home/server/%s'
        self.imageUrl = 'https://uploads.mangadex.org/data/%s/%%s'

    def starter(self):
        # Retrieve manga metadata from API
        mangaData = self.session.get(self.url)
        mangaData.raise_for_status()
        manga = mangaData.json()['data']

        # Retrieve chapter list from API
        chapterList = []
        chapterTotal = 1
        chapterOffset = 0
        while len(chapterList) < chapterTotal:
            chapterData = self.session.get(self.chaptersUrl + '&offset=%d' % chapterOffset)
            chapterData.raise_for_status()
            chapterBlock = chapterData.json()
            chapterTotal = chapterBlock['total']
            chapterOffset = chapterBlock['offset'] + chapterBlock['limit']
            chapterList.extend(chapterBlock['data'])

        # Determine if manga is complete and/or adult
        if manga['attributes']['lastChapter'] != '0':
            for chapter in chapterList:
                if chapter['attributes']['chapter'] == manga['attributes']['lastChapter']:
                    self.endOfLife = True
                    break

        if manga['attributes']['contentRating'] != 'safe':
            self.adult = True

        # Prepare chapter list
        self.chapters = []
        for chapter in chapterList:
            if chapter['attributes']['chapter'] == '':
                continue
            if chapter['attributes']['pages'] == 0:
                continue
            if len(self.chapters) >= 1 and chapter['attributes']['chapter'] == self.chapters[-1]['attributes']['chapter']:
                continue
            self.chapters.append(chapter)
        self.chapters.reverse()

        if not self.chapters:
            raise ValueError('Comic has no chapters!')

        # Find first and last chapter
        self.firstStripUrl = self.stripUrl % self.chapters[0]['id']
        return self.stripUrl % self.chapters[-1]['id']

    def getPrevUrl(self, url, data):
        # Determine previous chapter ID from cached list
        chapterId = util.urlpathsplit(url)[-1]
        chapter = list(filter(lambda c: c['id'] == chapterId, self.chapters))
        if len(chapter) == 0:
            return None
        return self.stripUrl % self.chapters[self.chapters.index(chapter[0]) - 1]['id']

    def extract_image_urls(self, url, data):
        # Retrieve chapter metadata from API
        chapters = json.loads(data.text_content())
        self.chapter = chapters['data']
        cdnresponse = self.session.get(self.cdnUrl % self.chapter['id'])
        cdnresponse.raise_for_status()
        cdnblock = cdnresponse.json()

        # Save link order for position-based filenames
        urltemplate = self.imageUrl % cdnblock['chapter']['hash']
        self._cached_image_urls = [urltemplate % page for page in cdnblock['chapter']['data']]
        return self._cached_image_urls

    def namer(self, imageUrl, pageUrl):
        # Construct filename from episode number and page index in array
        chapter = self.chapter['attributes']['chapter']
        chapterNum = chapter if chapter is not None else 0
        pageNum = self._cached_image_urls.index(imageUrl)
        pageExt = imageUrl.rsplit('.')[-1]
        return '%s-%02d.%s' % (chapterNum, pageNum, pageExt)

    @classmethod
    def getmodules(cls):
        return (
            cls('AniTomo', '920c22e7-49c9-4bb4-b394-0c964b6037fc'),
            cls('Beastars', 'f5e3baad-3cd4-427c-a2ec-ad7d776b370d'),
            cls('BokuNoKokoroNoYabaiYatsu', '3df1a9a3-a1be-47a3-9e90-9b3e55b1d0ac'),
            cls('CheerfulAmnesia', 'f9448f90-c068-4b6a-8c85-03d739aef255'),
            cls('DoChokkyuuKareshiXKanojo', 'efb62763-c940-4495-aba5-69c192a999a4'),
            cls('DragonDrive', '5c06ae70-b5cf-431a-bcd5-262a411de527'),
            cls('GanbareDoukiChan', '190616bc-7da6-45fd-abd4-dd2ca656c183'),
            cls('GijiHarem', 'd8f9afe2-aa44-4bc6-9145-eebb1f282372'),
            cls('HangingOutWithAGamerGirl', 'de9e3b62-eac5-4c0a-917d-ffccad694381'),
            cls('HoriMiya', 'a25e46ec-30f7-4db6-89df-cacbc1d9a900'),
            cls('HowToOpenATriangularRiceball', '6ebd90ce-d5e8-49c0-a4bc-e02e0f8ecb93'),
            cls('HunterXHunter', 'db692d58-4b13-4174-ae8c-30c515c0689c'),
            cls('IchaichasuruToOkaneGaWaichauFutariNoHanashi', '8eaaec7d-7aa7-490e-8d52-5a3d0a28e78b'),
            cls('InterspeciesReviewers', '1b2fddf9-1385-4f3c-b37a-cf86a9428b1a'),
            cls('JahySamaWaKujikenai', '2f4e5f5b-d930-4266-8c8a-c4cf9a81e51f'),
            cls('JingaiNoYomeToIchaIchaSuru', '809d2493-df3c-4e72-a57e-3e0026cae9fb'),
            cls('KawaiiJoushiWoKomarasetai', '23b7cc7a-df89-4049-af28-1fa78f88713e'),
            cls('KanojoOkarishimasu', '32fdfe9b-6e11-4a13-9e36-dcd8ea77b4e4'),
            cls('KoiToUtatane', 'f7d40a27-e289-45b3-9c68-d1cb251897e6'),
            cls('KonoKaishaNiSukiNaHitoGaImasu', '3e8cf40f-ba17-480a-b60b-a675db032ee2'),
            cls('Lv2KaraCheatDattaMotoYuushaKouhoNoMattariIsekaiLife', '58bc83a0-1808-484e-88b9-17e167469e23'),
            cls('MaouNoOreGaDoreiElfWoYomeNiShitandaGaDouMederebaIi', '55ace2fb-e157-4d76-9e72-67c6bd762a39'),
            cls('ModernMoGal', 'b1953f80-36f7-492c-b0f8-e9dd0ad01752'),
            cls('MousouTelepathy', '3d25d7da-893f-400e-9aeb-6163773c671a'),
            cls('MyTinySenpaiFromWork', '28ed63af-61f8-43af-bac3-762030c72963'),
            cls('MyWifeIsFromAThousandYearsAgo', '17a56d33-9443-433a-9e0d-70459893ed8f'),
            cls('OMaidensinYourSavageSeason', 'c4613b7d-7a6e-48f9-82f0-bce3dd33383a'),
            cls('OokamiShounenWaKyouMoUsoOKasaneru', '5e77d9e2-2e44-431a-a995-5fefd411e55e'),
            cls('OokamiToKoshinryou', 'de900fd3-c94c-4148-bbcb-ca56eaeb57a4'),
            cls('OtomeYoukaiZakuro', 'c1fa97be-0f1f-4686-84bc-806881c97d53'),
            cls('OversimplifiedSCP', 'e911fe33-a9b3-43dc-9eb7-f5ee081a6dc8'),
            cls('PleaseDontBullyMeNagatoroComicAnthology', '2a4bc9ec-2d70-428a-8b46-27f6218ed267'),
            cls('PleaseTellMeGalkochan', '7a2f2f6b-a6a6-4149-879b-3fc2f6916549'),
            cls('RebuildWorld', '99182618-ae92-4aec-a5df-518659b7b613'),
            cls('SeijoSamaIieToorisugariNoMamonotsukaiDesu', 'd4c40e73-251a-4bcb-a5a6-1edeec1e00e7'),
            cls('SenpaiGaUzaiKouhaiNoHanashi', 'af38f328-8df1-4b4c-a272-e737625c3ddc'),
            cls('SewayakiKitsuneNoSenkoSan', 'c26269c7-0f5d-4966-8cd5-b79acb86fb7a'),
            cls('ShinNoJitsuryokuWaGirigiriMadeKakushiteIyouToOmou', '22fda941-e603-4601-a536-c3ad6d004ba8'),
            cls('SwordArtOnline', '3dd0b814-23f4-4342-b75b-f206598534f6'),
            cls('SwordArtOnlineProgressive', '22ea3f54-11e4-4932-a527-89d63d3a62d9'),
            cls('TadokoroSan', '8ffbfa2f-23fa-4490-848e-942581a4d873'),
            cls('TamenDeGushi', '3f1453fb-9dac-4aca-a2ea-69613856c952'),
            cls('TheNewGate', 'b41bef1e-7df9-4255-bd82-ecf570fec566'),
            cls('TomoChanWaOnnanoko', '76ee7069-23b4-493c-bc44-34ccbf3051a8'),
            cls('TonikakuKawaii', '30f3ac69-21b6-45ad-a110-d011b7aaadaa'),
            cls('UramikoiKoiUramikoi', '009b6788-48f3-4e78-975c-097f54def7ab'),
            cls('UzakiChanWaAsobitai', '5a90308a-8b12-4a4d-9c6d-2487028fe319'),
            cls('YotsubaAnd', '58be6aa6-06cb-4ca5-bd20-f1392ce451fb'),
        )



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: © 2019 Tobias Gruetzmacher
# SPDX-FileCopyrightText: © 2019 Daniel Ring
from ..helpers import indirectStarter
from ..scraper import ParserScraper
from .common import WordPressScraper


class Housepets(WordPressScraper):
    url = 'https://www.housepetscomic.com/'
    stripUrl = url + 'comic/%s/'
    firstStripUrl = '2008/06/02/when-boredom-strikes'


class RickGriffinStudios(ParserScraper):
    imageSearch = '//img[@id="comic-image"]'
    prevSearch = '//a[@id="previous-button"]'
    latestSearch = '//a[@id="last-button"]'
    starter = indirectStarter
    nav = None

    def __init__(self, name, sub):
        super().__init__('RickGriffinStudios/' + name)
        self.stripUrl = 'https://rickgriffinstudios.com/' + sub + '/comic/%s/'
        self.url = self.firstStripUrl = self.stripUrl % '001'

    @classmethod
    def getmodules(cls):
        return (
            cls('AHClub', 'ah-club'),
            cls('HayvenCelestia', 'skinchange'),
            cls('TheStoryboard', 'the-storyboard'),
            cls('TheWitchOfKurikuto', 'the-witch-of-kurikuto'),
            cls('TracesOfThePast', 'traces-of-the-past'),
            cls('ZootopiaNightTerrors', 'zootopia-night-terrors'),
        )



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
# Copyright (C) 2004-2008 Tristan Seligmann and Jonathan Jacobs
# Copyright (C) 2012-2014 Bastian Kleineidam
# Copyright (C) 2015-2020 Tobias Gruetzmacher
'''update languages.py from pycountry'''
import os
import codecs

from dosagelib.scraper import scrapers


def main():
    """Update language information in dosagelib/languages.py."""
    basepath = os.path.dirname(os.path.dirname(__file__))
    fn = os.path.join(basepath, 'dosagelib', 'languages.py')
    with codecs.open(fn, 'w', 'utf-8') as f:
        f.write('# SPDX-License-Identifier: MIT\n')
        f.write('# ISO 693-1 language codes from pycountry\n')
        f.write('# This file is automatically generated, DO NOT EDIT!\n')
        lang = get_used_languages()
        write_languages(f, lang)


def get_used_languages():
    languages = {}
    for scraperobj in scrapers.get():
        lang = scraperobj.lang
        if lang not in languages:
            languages[lang] = scraperobj.language()
    return languages


def write_languages(f, langs):
    """Write language information."""
    f.write("Languages = {%s" % os.linesep)
    for lang in sorted(langs):
        f.write("    %r: %r,%s" % (lang, langs[lang], os.linesep))
    f.write("}%s" % os.linesep)


if __name__ == '__main__':
    main()



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: © 2004 Tristan Seligmann and Jonathan Jacobs
# SPDX-FileCopyrightText: © 2012 Bastian Kleineidam
# SPDX-FileCopyrightText: © 2015 Tobias Gruetzmacher
# PYTHON_ARGCOMPLETE_OK
from __future__ import annotations

import argparse
import contextlib
import importlib
import logging
import os
import platform
from collections.abc import Iterable

from platformdirs import PlatformDirs
from rich import columns, console, table, text

from . import AppName, __version__, configuration, director, events, output, singleton
from .scraper import scrapers as scrapercache
from .util import internal_error

logger = logging.getLogger(__name__)


class ArgumentParser(argparse.ArgumentParser):
    """Custom argument parser."""
    def __init__(self, console: console.Console) -> None:
        super().__init__(
            description="A comic downloader and archiver.",
            epilog=ExtraHelp,
            formatter_class=argparse.RawDescriptionHelpFormatter)
        self.console = console

    def print_help(self, file=None) -> None:
        """Paginate help message on TTYs."""
        with self.console.pager():
            self.console.print(self.format_help())


# Making our config roaming seems sensible
platformdirs = PlatformDirs(appname=AppName, appauthor=False, roaming=True, opinion=True)
user_plugin_path = platformdirs.user_data_path / 'plugins'


ExtraHelp = f"""\
EXAMPLES
List available comics:
  dosage -l

Get the latest comic of for example CalvinAndHobbes and save it in the "Comics"
directory:
  dosage CalvinAndHobbes

If you already have downloaded several comics and want to get the latest
strips of all of them:
  dosage --continue @

User plugin directory: {user_plugin_path}
"""


def setup_options(console: console.Console) -> ArgumentParser:
    """Construct option parser.
    @return: new option parser
    @rtype argparse.ArgumentParser
    """
    parser = ArgumentParser(console)

    parser.add_argument('-v', '--verbose', action='count', default=0,
        help='provides verbose output, use multiple times for more verbosity')
    parser.add_argument('-n', '--numstrips', action='store', type=int, default=0,
        help='traverse and retrieve the given number of comic strips;'
        ' use --all to retrieve all comic strips')
    parser.add_argument('-a', '--all', action='store_true',
        help='traverse and retrieve all comic strips')
    parser.add_argument('-c', '--continue', action='store_true', dest='cont',
        help='traverse and retrieve comic strips until an existing one is found')
    basepath_opt = parser.add_argument('-b', '--basepath', action='store',
        default='Comics', metavar='PATH',
        help='set the path to create invidivual comic directories in, default is Comics')
    parser.add_argument('--baseurl', action='store', metavar='PATH',
        help='the base URL of your comics directory (for RSS, HTML, etc.);'
        ' this should correspond to --base-path')
    parser.add_argument('-l', '--list', action='store_true',
        help='list available comic modules')
    parser.add_argument('--singlelist', action='store_true',
        help='list available comic modules in a single column list')
    parser.add_argument('--version', action='store_true',
        help='display the version number')
    parser.add_argument('--vote', action='store_true',
        help='vote for the selected comics')
    parser.add_argument('-m', '--modulehelp', action='store_true',
        help='display help for comic modules')
    parser.add_argument('-t', '--timestamps', action='store_true',
        help='print timestamps for all output at any info level')
    parser.add_argument('-o', '--output', action='append', dest='handler',
        choices=events.getHandlerNames(),
        help='sets output handlers for downloaded comics')
    parser.add_argument('--no-downscale', action='store_false',
        dest='allowdownscale',
        help='prevent downscaling when using html or rss handler')
    parser.add_argument('-p', '--parallel', action='store', type=int, default=1,
        help='fetch comics in parallel. Specify the number of connections')
    parser.add_argument('--adult', action='store_true',
        help='confirms that you are old enough to view adult content')
    parser.add_argument('--allow-multiple', action='store_true',
        help='allows multiple instances to run at the same time.'
        ' Use if you know what you are doing.')
    # used for development testing prev/next matching
    parser.add_argument('--dry-run', action='store_true',
        help=argparse.SUPPRESS)
    # List all comic modules, even those normally suppressed, because they
    # are not "real" (moved & removed)
    parser.add_argument('--list-all', action='store_true',
        help=argparse.SUPPRESS)
    comic_arg = parser.add_argument('comic', nargs='*',
        help='comic module name (including case insensitive substrings)')
    comic_arg.completer = scraper_completion
    with contextlib.suppress(ImportError):
        completers = importlib.import_module('argcomplete.completers')
        basepath_opt.completer = completers.DirectoriesCompleter()
        importlib.import_module('argcomplete').autocomplete(parser)
    return parser


def scraper_completion(**kwargs) -> Iterable[str]:
    """Completion helper for argcomplete."""
    scrapercache.adddir(user_plugin_path)
    return (comic.name for comic in scrapercache.all())


def display_version(verbose: bool) -> None:
    """Display application name, version, copyright and license."""
    print(configuration.App)
    print("Using Python {} ({}) on {}".format(platform.python_version(),
        platform.python_implementation(), platform.platform()))
    print(configuration.Copyright)
    print(configuration.Freeware)
    print("For support see", configuration.SupportUrl)
    if verbose:
        # search for updates
        from .updater import check_update
        try:
            value = check_update()
            if value:
                version, url = value
                if url is None:
                    # current version is newer than online version
                    text = ('Detected local or development version %(currentversion)s. '
                            'Available version of %(app)s is %(version)s.')
                else:
                    # display update link
                    text = ('A new version %(version)s of %(app)s is '
                            'available at %(url)s.')
                attrs = {'version': version, 'app': AppName,
                    'url': url, 'currentversion': __version__}
                print(text % attrs)
        except (IOError, KeyError) as err:
            print(f'An error occured while checking for an update of {AppName}: {err!r}')


def display_help(options) -> None:
    """Print help for comic strips."""
    for scraperobj in director.getScrapers(options.comic, options.basepath, listing=True):
        display_comic_help(scraperobj)


def display_comic_help(scraperobj) -> None:
    """Print help for a comic."""
    context = {"context": scraperobj.name}
    logger.info('URL: %s', scraperobj.url, extra=context)
    logger.info('Language: %s', scraperobj.language(), extra=context)
    if scraperobj.adult:
        logger.info("Adult comic, use option --adult to fetch.", extra=context)
    disabled = scraperobj.getDisabledReasons()
    if disabled:
        logger.info("Disabled: %s", " ".join(disabled.values()), extra=context)
    if scraperobj.help:
        logger.info(scraperobj.help, extra=context)


def vote_comics(options) -> int:
    """Vote for comics."""
    errors = 0
    for scraperobj in director.getScrapers(options.comic, options.basepath,
            options.adult):
        errors += vote_comic(scraperobj)
    return errors


def vote_comic(scraperobj) -> int:
    """Vote for given comic scraper."""
    context = {"context": scraperobj.name}
    try:
        scraperobj.vote()
        logger.info('Vote submitted.', extra=context)
    except Exception as msg:
        logger.exception(msg, extra=context)  # noqa: LOG010
        return 1
    return 0


def run(console: console.Console, options) -> int:
    """Execute comic commands."""
    err = 0
    output.console_logging(console, options.verbose, options.timestamps)
    scrapercache.adddir(user_plugin_path)
    # ensure only one instance of dosage is running
    if not options.allow_multiple:
        singleton.SingleInstance()

    if options.version:
        display_version(options.verbose)
    elif options.list:
        do_list(console)
    elif options.singlelist or options.list_all:
        do_list(console, column_list=False, verbose=options.verbose,
            listall=options.list_all)
    else:
        # after this a list of comic strips is needed
        if not options.comic:
            logger.warning('No comics specified, bailing out!')
            return 1

        if options.modulehelp:
            display_help(options)
        elif options.vote:
            err = vote_comics(options)
        else:
            err = director.getComics(options)
    return err


def do_list(console: console.Console, column_list=True, verbose=False, listall=False) -> None:
    """List available comics."""
    with console.pager():
        logger.info('Available comic scrapers:')
        logger.info('Comics tagged with [%s] require age confirmation'
            ' with the --adult option.', TAG_ADULT)
        logger.info('Non-english comics are tagged with [%s].', TAG_LANG)
        scrapers = sorted(scrapercache.all(listall),
                          key=lambda s: s.name.lower())
        if column_list:
            num, disabled = do_column_list(console, scrapers)
        else:
            num, disabled = do_single_list(console, scrapers, verbose=verbose)
        logger.info('%d supported comics.', num)
        if disabled:
            logger.info('')
            logger.info('Some comics are disabled, they are tagged with'
                ' [%s:REASON], where REASON is one of:', TAG_DISABLED)
            for k in disabled:
                logger.info('  %-10s %s', k, disabled[k])


def do_single_list(console: console.Console, scrapers, verbose=False):
    """Get list of scraper names, one per line."""
    disabled: dict[str, str] = {}
    for scraperobj in scrapers:
        if verbose:
            display_comic_help(scraperobj)
        else:
            tagged = get_tagged_scraper(scraperobj, reasons=disabled)
            console.print(text.Text(' '.join(filter(None, tagged))))
    return len(scrapers), disabled


def do_column_list(console: console.Console, scrapers):
    """Get list of scraper names with multiple names per line."""
    disabled: dict[str, str] = {}
    tagged_names = [get_tagged_scraper(scraperobj, reasons=disabled)
             for scraperobj in scrapers]
    lengths = sorted(len(tagged[0]) + len(tagged[1]) for tagged in tagged_names)
    upper_length = lengths[int(len(lengths) * 0.97) - 1]
    expectedcols = console.width // upper_length
    maxwidth = (console.width // expectedcols) - 3
    logger.debug("Calculated column widths: max: %i, upper: %i, width: %i",
        lengths[-1], upper_length, maxwidth)
    elements = (tagged_to_table(tagged, maxwidth) for tagged in tagged_names)
    console.print(columns.Columns(elements, equal=True, expand=True))
    return len(tagged_names), disabled


TAG_ADULT = "adult"
TAG_LANG = "lang"
TAG_DISABLED = "dis"


def get_tagged_scraper(scraperobj, reasons: dict[str, str]) -> tuple[str, str]:
    """Get comic scraper name."""
    tags = []
    if scraperobj.adult:
        tags.append(TAG_ADULT)
    if scraperobj.lang != "en":
        tags.append("%s:%s" % (TAG_LANG, scraperobj.lang))
    disabled = scraperobj.getDisabledReasons()
    if disabled and reasons is not None:
        reasons.update(disabled)
    for reason in disabled:
        tags.append("%s:%s" % (TAG_DISABLED, reason))

    return scraperobj.name, "[" + ", ".join(tags) + "]" if tags else ""


def tagged_to_table(tagged: tuple[str, str], limit: int) -> table.Table:
    output = table.Table.grid(padding=(0, 1))
    name, tags = tagged
    output.add_column(max_width=limit - len(tags), overflow="ellipsis")
    if tags:
        output.add_column()
        output.add_row(name, text.Text(tags))
    else:
        output.add_row(name)
    return output


def main(args=None):
    """Parse options and execute commands."""
    try:
        console = output.setup_console()
        argparser = setup_options(console)
        options = argparser.parse_args(args=args)
        options.basepath = os.path.expanduser(options.basepath)
        return run(console, options)
    except KeyboardInterrupt:
        print("Aborted.")
        return 1
    except Exception:
        internal_error()
        return 2



# SPDX-License-Identifier: MIT
# Copyright (C) 2019-2022 Tobias Gruetzmacher
import time
from pathlib import Path

import pytest


@pytest.fixture()
def _nosleep(monkeypatch):

    def sleep(seconds):
        pass

    monkeypatch.setattr(time, 'sleep', sleep)


@pytest.fixture()
def _noappdirs(monkeypatch):
    monkeypatch.setattr('dosagelib.cmd.user_plugin_path', Path(__file__).parent / 'mocks' / 'plugins')



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: © 2004 Tristan Seligmann and Jonathan Jacobs
# SPDX-FileCopyrightText: © 2012 Bastian Kleineidam
# SPDX-FileCopyrightText: © 2015 Tobias Gruetzmacher
# SPDX-FileCopyrightText: © 2019 Daniel Ring
from functools import cached_property
from re import MULTILINE, compile, escape

from .. import util
from ..helpers import bounceStarter, indirectStarter, joinPathPartsNamer
from ..scraper import ParserScraper, _BasicScraper, _ParserScraper
from ..util import tagre
from .common import (
    ComicControlScraper,
    WordPressNavi,
    WordPressScraper,
    WordPressSpliced,
    WordPressWebcomic,
)


class TekMage(WordPressNavi):
    url = 'https://tekmagecomic.com/'
    stripUrl = url + 'comic/%s/'
    firstStripUrl = stripUrl % 'chapter-1-page-1'


class Tamberlane(WordPressWebcomic):
    baseUrl = 'https://www.tamberlanecomic.com/'
    url = baseUrl + 'latest/'
    stripUrl = baseUrl + 'tamberlane/%s/'
    firstStripUrl = stripUrl % 'page-1'
    imageSearch = '//div[@id="comic-page"]/img/@src'
    prevSearch = '//a[@class="previous-link"]'


class TheBoyWhoFell(ComicControlScraper):
    url = 'https://www.boywhofell.com/'
    firstStripUrl = url + 'comic/ch00p00'


class TheBrads(_ParserScraper):
    url = ('https://web.archive.org/web/20171211154809/'
        'http://bradcolbow.com/archive/C4/')
    stripUrl = url + '%s/'
    firstStripUrl = stripUrl % 'P125'
    imageSearch = '//div[d:class("entry")]//img'
    prevSearch = '//a[d:class("prev")]'
    multipleImagesPerStrip = True
    endOfLife = True


class TheChroniclesOfHuxcyn(WordPressScraper):
    url = 'https://huxcyn.com/'
    stripUrl = url + 'comic/%s'
    firstStripUrl = stripUrl % 'opening-001'

    def namer(self, image_url, page_url):
        # Fix inconsistent filenames
        filename = util.urlpathsplit(image_url)[-1]
        filename = filename.replace('IMG_0504', 'TCoH109')
        filename = filename.replace('tcoh', 'TCoH')
        filename = filename.replace('1599151639.xizana_f3a6458e-8d94-4259-bec3-5a92706fe493_jpeg',
            'october.2020.cover')
        filename = filename.replace('huxonsword', 'october.2020.huxonsword')
        filename = filename.replace('New_Canvas100pageswebimage', 'TCoH100')
        if filename[0] == '0':
            filename = 'TCoH' + filename
        elif filename[0] == '3':
            pagenum = int(filename.rsplit('.', 1)[0].split('_', 1)[1].split('_', 1)[0])
            filename = 'TCoH' + str(40 + pagenum)
        return filename


class TheClassMenagerie(_ParserScraper):
    stripUrl = 'http://www.theclassm.com/d/%s.html'
    url = stripUrl % '20050717'
    firstStripUrl = stripUrl % '19990322'
    imageSearch = '//img[@class="ksc"]'
    prevSearch = '//a[@rel="prev"]'
    multipleImagesPerStrip = True
    endOfLife = True


class TheDepths(WordPressWebcomic):
    url = 'https://www.thedepthscomic.com/'
    stripUrl = url + 'comic/%s/'
    firstStripUrl = stripUrl % 'page-01'
    imageSearch = '//div[contains(@class, "webcomic-media")]//img'
    adult = True

    def namer(self, image_url, page_url):
        # Fix inconsistent filenames
        filename = util.urlpathsplit(image_url)[-1]
        filename = filename.replace('pg', 'page_')
        filename = filename.replace('page_', 'the_depths_')
        filename = filename.replace('-web', '')
        return filename


class TheDevilsPanties(WordPressNavi):
    url = 'https://thedevilspanties.com/'
    stripUrl = url + 'archives/%s'
    firstStripUrl = stripUrl % '300'
    help = 'Index format: number'


class TheDreamlandChronicles(WordPressScraper):
    url = 'http://www.thedreamlandchronicles.com/'


class TheForgottenOrder(ComicControlScraper):
    url = 'http://www.forgottenordercomic.com/'
    firstStripUrl = url + 'comic/prolouge-01-book-1'


class TheGamerCat(WordPressSpliced):
    url = 'https://thegamercat.com/'
    firstStripUrl = url + 'comic/06102011/'


class TheGentlemansArmchair(WordPressScraper):
    url = 'http://thegentlemansarmchair.com/'


class TheGentleWolf(WordPressScraper):
    url = 'https://thegentlewolf.net/'
    stripUrl = url + 'comic/%s/'
    firstStripUrl = stripUrl % 'tgw-001'

    def namer(self, image_url, page_url):
        # Fix duplicate filename
        filename = util.urlpathsplit(image_url)[-1]
        if page_url == self.stripUrl % 'tgw-271':
            filename = filename.replace('272', '271')
        return filename


class TheGlassScientists(ComicControlScraper):
    url = 'https://www.theglassscientists.com/'
    firstStripUrl = url + 'comic/chapter-i'


class TheJunkHyenasDiner(WordPressScraper):
    url = 'http://junkhyenasdiner.com/'
    stripUrl = url + 'comic/%s/'
    firstStripUrl = stripUrl % 'intro'


class TheLandscaper(_ParserScraper):
    stripUrl = ('https://web.archive.org/web/20171129163510/'
        'http://landscaper.visual-assault.net/comic/%s')
    url = stripUrl % 'latest'
    firstStripUrl = stripUrl % '1'
    imageSearch = '//article[d:class("comic")]//img[1]'
    prevSearch = '//a[contains(text(), "Previous")]'
    endOfLife = True


class TheMelvinChronicles(WordPressScraper):
    url = 'http://melvin.jeaniebottle.com/'


class TheNightBelongsToUs(_ParserScraper):
    url = 'https://tnbtu.com/'
    stripUrl = url + 'comic/%s/'
    firstStripUrl = stripUrl % '01-00'
    imageSearch = '//div[@id="spliced-comic"]//img'
    prevSearch = '//a[./img[contains(@src, "nav-prev")]]'
    latestSearch = '//a[contains(@class, "main-link")]'
    starter = indirectStarter
    adult = True


class TheNoob(WordPressScraper):
    url = 'http://thenoobcomic.com/'
    stripUrl = url + 'comic/%s/'
    firstStripUrl = stripUrl % '1'
    help = 'Index format: n (unpadded)'


class TheOldVictorian(ParserScraper):
    url = 'http://theoldvictorianwebcomic.com/'
    stripUrl = url + 'comic/%s/'
    firstStripUrl = stripUrl % 'the-old-victorian-cover'
    imageSearch = '//div[@id="comic"]//img'
    prevSearch = '//a[contains(@class, "comic-nav-previous")]'

    def namer(self, image_url, page_url):
        filename = util.urlpathsplit(image_url)[-1].replace('_', '-')
        filename = filename.replace('TOV00', 'TOV-00')
        if filename.replace('oldvic', '')[0].isdigit():
            filename = filename.replace('oldvic', 'TOV-00')
        if 'TOV-000' in filename and len(filename) > 12:
            filename = filename[:8] + '-' + filename[8:]
        return filename


class TheOrderOfTheStick(ParserScraper):
    url = 'https://www.giantitp.com/'
    stripUrl = url + 'comics/oots%s.html'
    firstStripUrl = stripUrl % '0001'
    imageSearch = '//img[contains(@src, "/comics/oots/")]'
    prevSearch = '//a[./img[@alt="Previous Comic"]]'
    latestSearch = '//a[@class="SideBar" and contains(@href, "/comics/oots")]'
    help = 'Index format: n (unpadded)'
    starter = indirectStarter
    namer = joinPathPartsNamer(pageparts=(-1,))


class TheRockCocks(ComicControlScraper):
    url = 'https://rockcocks.slipshine.net/'
    firstStripUrl = url + 'the-rock-cocks/page-1'
    adult = True


class TheWhiteboard(_ParserScraper):
    BROKEN_PAGE_MIDDLE = compile(r'</body></html>\n<')
    url = 'http://www.the-whiteboard.com/'
    stripUrl = url + 'auto%s.html'
    firstStripUrl = stripUrl % 'wb001'
    imageSearch = '//img[contains(@src, "auto")]'
    prevSearch = '//a[.//img[contains(@src, "previous")]]'

    def _parse_page(self, data):
        # Ugly hack to fix broken HTML
        data = self.BROKEN_PAGE_MIDDLE.sub('<', data)
        return super(TheWhiteboard, self)._parse_page(data)

    def imageUrlModifier(self, url, data):
        return self.url + url

    def link_modifier(self, fromurl, tourl):
        return self.url + tourl


class TheWotch(WordPressScraper):
    url = 'http://www.thewotch.com/'
    firstStripUrl = url + '?comic=enter-the-wotch'


class ThisIsIndexed(_BasicScraper):
    url = 'http://thisisindexed.com/'
    rurl = escape(url)
    stripUrl = url + 'page/%s'
    imageSearch = compile(tagre("img", "src", r'(%swp-content/uploads/\d+/\d+/card[^"]+)' % rurl))
    multipleImagesPerStrip = True
    prevSearch = compile(tagre("div", "class", "nav-previous") +
                         tagre("a", "href", r'(%spage/\d+/)[^"]*' % rurl))
    help = 'Index format: number'


class ThreePanelSoul(ComicControlScraper):
    url = 'http://threepanelsoul.com/'
    firstStripUrl = url + 'comic/a-test-comic'


class TinyDickAdventures(ParserScraper):
    url = 'https://www.lfg.co/'
    stripUrl = url + 'tda/strip/%s/'
    firstStripUrl = stripUrl % '1'
    imageSearch = '//div[@id="comic-img"]//img'
    prevSearch = '//a[@class="comic-nav-prev"]'
    latestSearch = '//div[@id="feature-tda-footer"]/a[contains(@href, "tda/strip/")]'
    starter = indirectStarter
    namer = joinPathPartsNamer(pageparts=(-1,))


class ToonHole(ParserScraper):
    url = 'https://toonhole.com/'
    firstStripUrl = url + '2010/01/smart-questions-get-smart-answers/'
    imageSearch = '//img[d:class("wp-post-image")]'
    prevSearch = '//a[d:class("previous-post")]'
    latestSearch = '//h2[d:class("entry-title")]/a'
    starter = indirectStarter
    namer = joinPathPartsNamer(imageparts=range(-3, 0))


class TrippingOverYou(_BasicScraper):
    url = 'http://www.trippingoveryou.com/'
    stripUrl = url + 'comic/%s'
    firstStripUrl = stripUrl % 'wiggle-room'
    imageSearch = compile(tagre("img", "src", r'([^"]+/comics/[^"]+)'))
    prevSearch = compile(r'<a class="cc-prev" rel="prev" href="(.+?)">')
    help = 'Index format: stripname'


class TumbleDryComics(WordPressScraper):
    url = 'https://www.tumbledrycomics.com/'
    stripUrl = url + 'comic/%s/'
    firstStripUrl = stripUrl % 'we-need-to-get-high-jpg'
    textSearch = '//div[@id="comic"]//img/@alt'
    multipleImagesPerStrip = True
    adult = True
    help = 'Index format: name'

    def namer(self, image_url, page_url):
        # Most images have the date they were posted in the filename
        # For those that don't we can get the month and year from the image url
        year, month, filename = util.urlpathsplit(image_url)[-3:]
        if not filename.startswith(year):
            filename = year + "-" + month + "-" + filename
        return filename


class Turnoff(ParserScraper):
    name = 'turnoff'
    url = 'https://turnoff.us/'
    imageSearch = '//article[d:class("post-content")]//img'
    prevSearch = '//div[d:class("prev")]//a'
    nextSearch = '//a[text()="next"]'
    stripUrl = url + 'geek/%s/'
    firstStripUrl = stripUrl % 'tcp-buddies'
    multipleImagesPerStrip = True
    starter = bounceStarter

    @cached_property
    def comics_order(self):
        # Neither the images nor the pages contain information about dates or indices.
        # However we can extract the order of the images from the JavaScript.
        html = self.session.get(self.url).text
        list_regex = compile(r"""^\s*"/geek/(.*?)/?",\s*$""", flags=MULTILINE)
        return list(reversed(list_regex.findall(html)))

    def namer(self, image_url, page_url):
        comic_name = util.urlpathsplit(page_url)[-1]
        index = self.comics_order.index(comic_name) + 1
        file_name = util.urlpathsplit(image_url)[-1]
        return "%03d-%s" % (index, file_name)


class TwinDragons(WordPressScraper):
    url = 'http://www.twindragonscomic.com/'
    stripUrl = url + 'comic/%s/'
    firstStripUrl = stripUrl % 'the-beginning'
    multipleImagesPerStrip = True


class TwoGuysAndGuy(ComicControlScraper):
    url = ('https://web.archive.org/web/20230720180648/'
        'https://twogag.com/')
    stripUrl = url + 'archives/%s'
    firstStripUrl = stripUrl % '4'
    help = 'Index format: number[-name]'
    adult = True
    endOfLife = True


class Twokinds(_ParserScraper):
    url = 'http://twokinds.keenspot.com/'
    stripUrl = url + 'comic/%s/'
    firstStripUrl = stripUrl % '1'
    imageSearch = '//article[d:class("comic")]//img'
    prevSearch = '//a[d:class("navprev")]'
    help = 'Index format: n (unpadded)'


class TwokindsSketches(Twokinds):
    name = 'Twokinds/Sketches'
    imageSearch = '//article[contains(@class, "comic")]/a'


class TwoLumps(_BasicScraper):
    url = 'http://www.twolumps.net/'
    stripUrl = url + 'd/%s.html'
    imageSearch = compile(tagre("img", "src", r'(/comics/[^"]+)'))
    prevSearch = compile(tagre("a", "href", r'(/d/\d+\.html)', after="prev"))
    help = 'Index format: yyyymmdd'



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: © 2004 Tristan Seligmann and Jonathan Jacobs
# SPDX-FileCopyrightText: © 2012 Bastian Kleineidam
# SPDX-FileCopyrightText: © 2015 Tobias Gruetzmacher
import pytest
import requests
import responses

from dosagelib import cmd, scraper


class ATestScraper(scraper.BasicScraper):
    pass


class TestVote(object):
    @responses.activate
    def test_vote(self):
        responses.add(responses.POST, 'https://buildbox.23.gs/count/')

        ATestScraper('Test_Test').vote()

    @responses.activate
    def test_vote_err(self):
        responses.add(responses.POST, 'https://buildbox.23.gs/count/', status=500)

        with pytest.raises(requests.exceptions.HTTPError):
            ATestScraper('Test_Test').vote()

    @responses.activate
    def test_run_vote(self):
        responses.add(responses.POST, 'https://buildbox.23.gs/count/')

        assert cmd.main(('-v', '--vote', 'xkcd')) == 0

    @responses.activate
    def test_run_vote_err(self):
        responses.add(responses.POST, 'https://buildbox.23.gs/count/', status=500)

        assert cmd.main(('-v', '--allow-multiple', '--vote', 'xkcd')) == 1



# SPDX-License-Identifier: MIT
# Copied from: https://github.com/pycontribs/tendo
# License: PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# Author: Sorin Sbarnea
# Changes: changed logging and formatting
from dosagelib import singleton
from multiprocessing import Process


def f(flavor_id):
    return singleton.SingleInstance(flavor_id=flavor_id, exit_code=1)


class TestSingleton(object):
    def test_1(self):
        # test in current process
        me = singleton.SingleInstance(flavor_id="test-1")
        del me  # now the lock should be removed
        assert True

    def test_2(self):
        # test in current subprocess
        p = Process(target=f, args=("test-2",))
        p.start()
        p.join()
        # the called function should succeed
        assert p.exitcode == 0

    def test_3(self):
        # test in current process and subprocess with failure
        # start first instance
        me = f("test-3")
        # second instance
        p = Process(target=f, args=("test-3",))
        p.start()
        p.join()
        assert p.exitcode == 1
        # third instance
        p = Process(target=f, args=("test-3",))
        p.start()
        p.join()
        assert p.exitcode == 1
        del me  # now the lock should be removed



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: © 2017 Tobias Gruetzmacher
import gzip
import os.path
import re
from functools import lru_cache

from responses import GET, add


def _file(name):
    return os.path.join(os.path.dirname(__file__), 'responses', name)


@lru_cache()
def content(name, ext: str = 'html'):
    with gzip.open(_file(f'{name}.{ext}.gz'), 'r') as f:
        return f.read()


@lru_cache()
def _img(name):
    with open(_file(name + '.png'), 'rb') as f:
        return f.read()


def page(url, pagename):
    add(GET, url, content(pagename))


def json(url, pagename):
    add(GET, url, content(pagename, 'json'))


def png(url, name='empty'):
    add(GET, url, _img(name), content_type='image/png')


def jpeg(url, name='empty'):
    add(GET, url, _img(name), content_type='image/jpeg')


def xkcd():
    page('https://xkcd.com/', 'xkcd-1899')
    for num in (302, 303, 1898, 1899):
        page('https://xkcd.com/%i/' % num, 'xkcd-%i' % num)
    png(re.compile(r'https://imgs\.xkcd\.com/.*\.png'))



# SPDX-License-Identifier: PSF-2.0
# Copied from: https://github.com/pycontribs/tendo
# Author: Sorin Sbarnea
# Changes: changed logging and formatting
import errno
import logging
import os
import sys
import tempfile

logger = logging.getLogger(__name__)


class SingleInstance:
    """
    To prevent a script from running in parallel instantiate the
    SingleInstance() class. If is there another instance
    already running it will exit the application with the message
    "Another instance is already running, quitting.",
    returning an error code of -1.

    >>> me = SingleInstance()

    This is very useful to execute scripts by crontab that should only run
    one at a time.

    Note that this works by creating a lock file with a filename based
    on the full path to the script file.
    """

    def __init__(self, flavor_id="", exit_code=-1):
        """Create an exclusive lockfile or exit with an error and the given
        exit code."""
        self.initialized = False
        scriptname = os.path.splitext(os.path.realpath(sys.argv[0]))[0]
        lockname = scriptname.replace("/", "-").replace(":", "").replace("\\", "-")
        if flavor_id:
            lockname += "-%s" % flavor_id
        lockname += '.lock'
        tempdir = tempfile.gettempdir()
        self.lockfile = os.path.normpath(os.path.join(tempdir, lockname))
        logger.debug("SingleInstance lockfile: %s", self.lockfile)
        if sys.platform == 'win32':
            try:
                # file already exists, try to remove it in case the previous
                # execution was interrupted
                if os.path.exists(self.lockfile):
                    os.unlink(self.lockfile)
                self.fd = os.open(self.lockfile, os.O_CREAT | os.O_EXCL | os.O_RDWR)
            except OSError:
                type, e, tb = sys.exc_info()
                if e.errno == errno.EACCES:  # EACCES == 13
                    self.exit(exit_code)
                raise
        else:  # non Windows
            import fcntl
            self.fp = open(self.lockfile, 'w')
            try:
                fcntl.lockf(self.fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
                # raises IOError on Python << 3.3, else OSError
            except OSError:
                self.exit(exit_code)
        self.initialized = True

    def exit(self, exit_code):
        """Exit with an error message and the given exit code."""
        logger.error("Another instance is already running, quitting.")
        sys.exit(exit_code)

    def __del__(self):
        """Remove the lock file."""
        if not self.initialized:
            return
        try:
            if sys.platform == 'win32':
                if hasattr(self, 'fd'):
                    os.close(self.fd)
                    os.unlink(self.lockfile)
            else:
                import fcntl
                fcntl.lockf(self.fp, fcntl.LOCK_UN)
                if os.path.isfile(self.lockfile):
                    os.unlink(self.lockfile)
        except Exception:
            logger.exception("could not remove lockfile")



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: © 2019 Tobias Gruetzmacher
# SPDX-FileCopyrightText: © 2019 Thomas W. Littauer
"""
Script to get a list of comicskingdom.com comics and save the info in a JSON
file for further processing.
"""

from dosagelib import util
from scriptutil import ComicListUpdater


class ComicsKingdomUpdater(ComicListUpdater):
    dup_templates = (
        "Creators/%s",
        "GoComics/%s",
        "KeenSpot/%s",
        "ComicGenesis/%s",
    )

    def handle_listing(self, page):
        for link in self.xpath(page, '//ul[d:class("index")]//a'):
            name = link.text_content().removeprefix('The ')
            url = link.attrib['href']
            lang = 'es' if ' (Spanish)' in name else None

            self.add_comic(name, (url, lang))

    def collect_results(self):
        """Parse all search result pages."""
        self.handle_listing(self.get_url('https://comicskingdom.com/features'))

    def get_entry(self, name: str, data: tuple[str, str]):
        opt = f", lang='{data[1]}'" if data[1] else ''
        pathparts = util.urlpathsplit(data[0])
        path = pathparts[1] if pathparts[0] == 'vintage' else pathparts[0]
        return f"cls('{name}', '{path}'{opt}),"


if __name__ == '__main__':
    ComicsKingdomUpdater(__file__).run()



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: © 2004 Tristan Seligmann and Jonathan Jacobs
# SPDX-FileCopyrightText: © 2012 Bastian Kleineidam
# SPDX-FileCopyrightText: © 2015 Tobias Gruetzmacher
# SPDX-FileCopyrightText: © 2019 Daniel Ring
from re import compile, escape

from .. import util
from ..helpers import bounceStarter, indirectStarter, joinPathPartsNamer
from ..scraper import ParserScraper, _BasicScraper, _ParserScraper
from ..util import tagre
from .common import ComicControlScraper, WordPressNavi, WordPressScraper


class PandyLand(_ParserScraper):
    url = ('https://web.archive.org/web/20200122163307/'
        'http:/pandyland.net/')
    stripUrl = url + '%s'
    firstStripUrl = stripUrl % '1'
    imageSearch = '//div[d:class("comic")]/img'
    prevSearch = '//a[contains(text(), "previous")]'
    help = 'Index format: number'
    endOfLife = True


class ParadigmShift(_BasicScraper):
    url = 'http://www.paradigmshiftmanga.com/'
    starter = indirectStarter
    stripUrl = url + 'ps/%s.html'
    imageSearch = compile(tagre("img", "src", r'([^"]*comics/ps/[^"]*)'))
    prevSearch = compile(tagre("a", "href", r'([^"]+)',
                               after="previous-comic-link"))
    latestSearch = compile(tagre("a", "href", r'([^"]+)',
                                 after="next-comic-link"))
    help = 'Index format: custom'


class ParallelUniversum(WordPressScraper):
    url = 'https://www.paralleluniversum.net/'
    stripUrl = url + '%s/'
    firstStripUrl = stripUrl % '001-der-comic-ist-tot'
    prevSearch = '//a[@rel="prev"]'
    lang = 'de'


class ParaNatural(ComicControlScraper):
    url = 'https://www.paranatural.net/'
    firstStripUrl = url + 'comic/chapter-1'


class PartiallyClips(WordPressScraper):
    url = ('https://web.archive.org/web/20180509161332/'
        'http://partiallyclips.com/')
    firstStripUrl = url + 'comic/screaming-woman/'
    endOfLife = True


class PastelDefender(_BasicScraper):
    baseUrl = 'http://www.pasteldefender.com/'
    url = baseUrl + 'coverbackcover.html'
    stripUrl = baseUrl + '%s.html'
    firstStripUrl = stripUrl % 'cover'
    imageSearch = compile(r'<IMG SRC="(images/.+?)" WIDTH="742"')
    prevSearch = compile(r'<A HREF="([^"]+)"><IMG SRC="images/back\.gif"')
    help = 'Index format: nnn'


class PeanutBerrySundae(_ParserScraper):
    baseUrl = 'https://foxyverse.com/'
    url = baseUrl + 'comics/'
    stripUrl = baseUrl + 'peanut-berry-sundae-%s/'
    firstStripUrl = stripUrl % '1'
    imageSearch = ('//img[contains(@src, "Page")]',
                   '//img[contains(@src, "page")]')
    latestSearch = '//a[contains(@href, "peanut-berry-sundae")]'
    starter = indirectStarter
    adult = True

    def getPrevUrl(self, url, data):
        # Replace missing navigation links
        pageNum = int(url.replace('-70-2', '-71').rstrip('/').rsplit('-', 1)[-1])
        url = self.stripUrl % str(pageNum - 1)
        return url.replace('-71', '-70-2')


class PebbleVersion(_ParserScraper):
    url = 'http://www.pebbleversion.com/'
    stripUrl = url + 'Archives/Strip%s.html'
    imageSearch = "//table/tr[2]//img"
    prevSearch = '//a[text()="Previous Comic"]'
    help = 'Index format: n (unpadded)'


class PennyAndAggie(ComicControlScraper):
    url = 'https://pixietrixcomix.com/penny-and-aggie'
    stripUrl = url + '/%s'
    firstStripUrl = stripUrl % '2004-09-06'
    endOfLife = True
    help = 'Index format: yyyy-mm-dd'


class PennyArcade(_ParserScraper):
    url = 'https://www.penny-arcade.com/comic/'
    stripUrl = url + '%s'
    firstStripUrl = stripUrl % '1998/11/18/the-sin-of-long-load-times'
    imageSearch = '//div[d:class("comic-panel")]//img'
    prevSearch = '//a[d:class("older")]'
    nextSearch = '//a[d:class("newer")]'
    multipleImagesPerStrip = True
    starter = bounceStarter
    help = 'Index format: yyyy/mm/dd'


class PeppermintSaga(WordPressScraper):
    url = 'http://www.pepsaga.com/'
    stripUrl = url + 'comics/%s/'
    firstStripUrl = stripUrl % 'the-sword-of-truth-vol1'
    adult = True


class PeppermintSagaBGR(WordPressScraper):
    url = 'http://bgr.pepsaga.com/'
    stripUrl = url + '?comic=%s'
    firstStripUrl = stripUrl % '04172011'
    adult = True


class PeterAndCompany(_ParserScraper):
    url = 'http://peterandcompany.com/'
    stripUrl = url + '%s'
    firstStripUrl = stripUrl % '20050101'
    imageSearch = ('//div[@id="page"]//img',
                   '//div[@id="strip"]//img[contains(@src, "strips/")]')
    prevSearch = '//a[./img[contains(@src, "nav_previous")]]'


class PeterAndWhitney(_ParserScraper):
    url = 'http://peterandwhitney.com/'
    stripUrl = url + '%s'
    firstStripUrl = stripUrl % '20160502'
    imageSearch = '//div[@id="page"]//img'
    prevSearch = '//a[./img[contains(@src, "nav_previous")]]'


class PHDComics(ParserScraper):
    BROKEN_COMMENT_END = compile(r'--!>')

    baseUrl = 'https://phdcomics.com/'
    url = baseUrl + 'comics.php'
    stripUrl = baseUrl + 'comics/archive.php?comicid=%s'
    firstStripUrl = stripUrl % '1'
    imageSearch = (r'//img[starts-with(@name, "comic") and re:test(@src, "\.[a-z]{3,4}$")]',
        '//img[d:class("img-responsive") and @title]')
    multipleImagesPerStrip = True
    prevSearch = '//a[img[contains(@src, "prev_button")]]'
    nextSearch = '//a[img[contains(@src, "next_button")]]'
    help = 'Index format: n (unpadded)'

    # Ugly hack :(
    def _parse_page(self, data):
        data = self.BROKEN_COMMENT_END.sub('-->', data)
        return super(PHDComics, self)._parse_page(data)

    def shouldSkipUrl(self, url, data):
        """Skip pages without images."""
        return url in (
            # video
            self.stripUrl % '1880',
            self.stripUrl % '1669',
        ) or self.match(data, '//img[@id="comic" and contains(@src, "phd083123s")]')


class Picklewhistle(ComicControlScraper):
    url = 'http://www.picklewhistle.com/'


class PicPakDog(WordPressScraper):
    url = 'http://www.picpak.net/'
    firstStripUrl = url + 'comic/dogs-cant-spell/'


class Pixel(_BasicScraper):
    url = 'http://pixelcomic.net/'
    rurl = escape(url)
    stripUrl = url + '%s'
    firstStripUrl = stripUrl % '000.shtml'
    imageSearch = compile(tagre("img", "src", r'(\d+\.png)'))
    prevSearch = compile(tagre("a", "href", r'(%s\d+\.(?:php|shtml))' % rurl,
                               before="prev"))
    help = 'Index format: nnn'


class PlanescapeSurvival(_BasicScraper):
    url = 'http://planescapecomic.com/'
    stripUrl = url + '%s.html'
    imageSearch = compile(r'src="(comics/.+?)"')
    prevSearch = compile(r'<a href="(.+?)"><img alt="Previous" ')
    help = 'Index format: nnn'


class PlushAndBlood(ParserScraper):
    url = 'http://www.plushandblood.com/Comic.php'
    stripUrl = url + '?strip_id=%s'
    firstStripUrl = stripUrl % '1'
    imageSearch = '//img[contains(@src, "comics/")]'
    prevSearch = '//a[contains(text(), "PREV")]'


class PokeyThePenguin(ParserScraper):
    url = 'https://www.yellow5.com/pokey/archive/'
    stripUrl = url + 'index%s.html'
    firstStripUrl = stripUrl % '1'
    imageSearch = '//p/img'
    latestSearch = '(//a)[last()]'
    multipleImagesPerStrip = True
    starter = indirectStarter
    help = 'Index format: number'

    def getPrevUrl(self, url, data):
        """Decrease index.html number."""
        mo = compile(r"index(\d+)\.html").search(url)
        num = int(mo.group(1)) - 1
        prefix = url.rsplit('/', 1)[0]
        return f"{prefix}/index{num}.html"


class PoorlyDrawnLines(ParserScraper):
    url = 'https://poorlydrawnlines.com/'
    stripUrl = url + 'comic/%s/'
    multipleImagesPerStrip = True
    firstStripUrl = stripUrl % 'hardly-essayists'
    imageSearch = '//div[d:class("entry-content")]//noscript/img'
    prevSearch = '//a[@rel="prev"]'
    namer = joinPathPartsNamer(imageparts=range(-3, 0))

    def shouldSkipUrl(self, url, _data):
        """Skip pages without a comic."""
        skipUrls = [self.stripUrl % s for s in (
            'hope-it-all-works-out-new-book-coming-this-fall',
            'poorly-drawn-lines-animated-series',
            'poorly-drawn-lines-episode-two',
            'watch-poorly-drawn-lines-on-hulu',
        )]
        return url in skipUrls

    def getPrevUrl(self, url: str, data):
        """Skip missing comics which redirect back to home page"""
        if url == self.stripUrl % '8198':
            return self.stripUrl % 'excited-2'
        elif url == self.stripUrl % '8186':
            return self.stripUrl % 'to-hell-2'
        elif url == self.stripUrl % '8177':
            return self.stripUrl % 'feel-real'
        elif url == self.stripUrl % '2056':
            return self.stripUrl % 'stereotype'

        return super().getPrevUrl(url, data)


class PoppyOPossum(WordPressScraper):
    baseUrl = 'https://www.poppy-opossum.com/'
    url = baseUrl + '?latest'
    stripUrl = baseUrl + 'comic/%s'
    firstStripUrl = stripUrl % 'a-story'


class PowerNap(_ParserScraper):
    url = 'https://www.powernapcomic.com/powernap/'
    stripUrl = url + 'd/%s.html'
    firstStripUrl = stripUrl % '20110617'
    imageSearch = '//center/img'
    prevSearch = '//a[./img[contains(@src, "previous")]]'
    endOfLife = True

    def imageUrlModifier(self, url, data):
        return url.replace('\n', '').strip()

    def link_modifier(self, fromurl, tourl):
        return tourl.replace('\n', '').strip()


class Precocious(_ParserScraper):
    url = 'http://www.precociouscomic.com/'
    stripUrl = url + 'archive/comic/%s'
    firstStripUrl = stripUrl % '2009/03/09'
    imageSearch = '//img[contains(@src, "/comics/")]'
    prevSearch = '//a[img[contains(@src, "/back_arrow")]]'
    help = 'Index format: yyyy/mm/dd'


class PrinceOfSartar(WordPressNavi):
    url = 'http://www.princeofsartar.com/'
    stripUrl = url + 'comic/%s/'
    firstStripUrl = stripUrl % 'introduction-chapter-1'
    nextSearch = '//a[d:class("navi-next")]'
    starter = bounceStarter
    namer = joinPathPartsNamer(pageparts=(-1,))
    help = 'Index format: name'


class ProphecyOfTheCircle(WordPressNavi):
    url = 'https://www.prophecyofthecircle.com/'
    stripUrl = url + 'comic/%s/'
    firstStripUrl = stripUrl % 'prologue'

    def namer(self, image_url, page_url):
        # Fix duplicate filenames
        filename = util.urlpathsplit(image_url)[-1]
        if '2015/12/jamet101' in image_url:
            filename = filename.replace('101', '10')
        elif '2012-02-20-jahrd156' in image_url:
            filename = filename.replace('156', '157')
        elif '2011-10-02-jahrd137' in image_url:
            filename = filename.replace('137', '137-1')
        # Fix inconsistent filenames
        if filename[0] == '2':
            filename = filename[11:]
        if len(filename) >= 9 and filename[8].isdigit():
            filename = filename[:8] + '-' + filename[8:]
        return filename


class Prototype(_ParserScraper):
    stripUrl = 'https://web.archive.org/web/20201030035444/http://planetprototype.com/%s/'
    firstStripUrl = stripUrl % '2018/03/30/vol-1-ch-1-front-cover'
    url = firstStripUrl
    imageSearch = '//img[contains(@class, "wp-post-image")]'
    prevSearch = '//a[.//text()="Previous"]'
    latestSearch = '//a[.//text()="Latest"]'
    starter = indirectStarter
    endOfLife = True


class PS238(_ParserScraper):
    url = 'http://ps238.nodwick.com/'
    stripUrl = url + 'comic/%s/'
    firstStripUrl = stripUrl % '12072006'
    imageSearch = '//div[@id="comic"]//img'
    prevSearch = '//a[@class="comic-nav-base comic-nav-previous"]'
    help = 'Index format: yyyy-mm-dd'


class PvPOnline(ParserScraper):
    baseUrl = 'https://www.toonhoundstudios.com/'
    stripUrl = baseUrl + 'comic/%s/?sid=372'
    url = stripUrl % 'pvp-2022-09-16'
    firstStripUrl = stripUrl % '19980504'
    imageSearch = '//div[@id="spliced-comic"]//img/@data-src-img'
    prevSearch = '//a[d:class("prev")]'
    endOfLife = True

    def namer(self, image_url, page_url):
        return 'pvp' + util.urlpathsplit(image_url)[-1]



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: © 2004 Tristan Seligmann and Jonathan Jacobs
# SPDX-FileCopyrightText: © 2012 Bastian Kleineidam
# SPDX-FileCopyrightText: © 2015 Tobias Gruetzmacher
from __future__ import annotations

import codecs
import contextlib
import glob
import logging
import mimetypes
import os
from datetime import datetime
from typing import Iterator

from rich import filesize

from .events import getHandler
from .util import getFilename, unquote, urlopen

logger = logging.getLogger(__name__)

# Maximum content size for images
MAX_IMAGE_BYTES = 1024 * 1024 * 20  # 20 MB
# RFC 1123 format, as preferred by RFC 2616
RFC_1123_DT_STR = "%a, %d %b %Y %H:%M:%S GMT"


class ComicStrip:
    """A list of comic image URLs."""

    def __init__(self, scraper, strip_url: str, image_urls: str, text=None) -> None:
        """Store the image URL list."""
        self.scraper = scraper
        self.strip_url = strip_url
        self.image_urls = image_urls
        self.text = text

    def getImages(self) -> Iterator[ComicImage]:
        """Get a list of image downloaders."""
        for image_url in self.image_urls:
            yield self.getDownloader(image_url)

    def getDownloader(self, url: str) -> ComicImage:
        """Get an image downloader."""
        filename = self.scraper.namer(url, self.strip_url)
        return ComicImage(self.scraper, url, self.strip_url, filename,
                          text=self.text)


class ComicImage:
    """A comic image downloader."""

    ChunkBytes = 1024 * 100  # 100KB

    def __init__(self, scraper, url, referrer, filename, text=None) -> None:
        """Set URL and filename."""
        self.scraper = scraper
        self.referrer = referrer
        self.url = url
        filename = getFilename(filename)
        self.filename, self.ext = os.path.splitext(filename)
        if not self.ext:
            self.ext = '.bin'
        self.text = text

    def connect(self, lastchange=None):
        """Connect to host and get meta information."""
        headers = {}
        if lastchange:
            headers['If-Modified-Since'] = lastchange.strftime(RFC_1123_DT_STR)
        self.urlobj = urlopen(self.url, self.scraper.session,
                              referrer=self.referrer,
                              max_content_bytes=MAX_IMAGE_BYTES, stream=True,
                              headers=headers)
        if self.urlobj.status_code == 304:  # Not modified
            return
        content_type = unquote(self.urlobj.headers.get(
            'content-type', 'application/octet-stream'))
        content_type = content_type.split(';', 1)[0]
        maintype = content_type.split('/', 1)[0]
        if maintype != 'image' and content_type not in (
                'application/octet-stream', 'application/x-shockwave-flash'):
            raise IOError('content type %r is not an image at %s' % (
                content_type, self.url))
        # Try to guess "better" extension from mime type
        guessed_ext = mimetypes.guess_extension(content_type)
        if guessed_ext and guessed_ext != '.bin':
            self.ext = guessed_ext
        self.contentLength = int(self.urlobj.headers.get('content-length', 0))
        logger.debug('... filename = %r, ext = %r, contentLength = %d', self.filename, self.ext, self.contentLength)

    def save(self, basepath):
        """Save comic URL to filename on disk."""
        fnbase = self._fnbase(basepath)
        exist = [x for x in glob.glob(fnbase + ".*") if not x.endswith(".txt")]
        logger.moreinfo("Get image URL %r", self.url)
        if len(exist) == 1:
            lastchange = os.path.getmtime(exist[0])
            self.connect(datetime.utcfromtimestamp(lastchange))
            if self.urlobj.status_code == 304:  # Not modified
                self._exist_err(exist[0])
                return exist[0], False
        else:
            self.connect()
        fn = fnbase + self.ext
        # compare with >= since content length could be the compressed size
        if os.path.isfile(fn) and os.path.getsize(fn) >= self.contentLength:
            self._exist_err(fn)
            return fn, False
        logger.debug('Writing comic to file %r...', fn)
        with self.fileout(fn) as f:
            for chunk in self.urlobj.iter_content(self.ChunkBytes):
                f.write(chunk)
        if self.text:
            fntext = fnbase + ".txt"
            logger.debug('Writing comic text to file %s...', fntext)
            with self.fileout(fntext, encoding='utf-8') as f:
                f.write(self.text)
        getHandler().comicDownloaded(self, fn)
        return fn, True

    @contextlib.contextmanager
    def fileout(self, filename, encoding=None):
        """Write content to given filename. Checks for zero-sized files.
        If encoding is given writes to a codec.open() file."""
        def getfp(filename, encoding):
            """Get open file object."""
            if encoding:
                return codecs.open(filename, 'w', encoding)
            return open(filename, 'wb')

        try:
            with getfp(filename, encoding) as fp:
                yield fp
                size = fp.tell()
        except Exception:
            if os.path.isfile(filename):
                os.remove(filename)
            raise
        else:
            logger.info("Saved %r (%s).", filename, filesize.decimal(size))

    def _exist_err(self, fn):
        logger.info('Skipping existing file %r.', fn)

    def _fnbase(self, basepath):
        '''Determine the target base name of this comic file and make sure the
        directory exists.'''
        comicdir = self.scraper.get_download_dir(basepath)
        if not os.path.isdir(comicdir):
            os.makedirs(comicdir)
        return os.path.join(comicdir, self.filename)



# SPDX-License-Identifier: MIT
# Copyright (C) 2004-2008 Tristan Seligmann and Jonathan Jacobs
# Copyright (C) 2012-2014 Bastian Kleineidam
# Copyright (C) 2015-2022 Tobias Gruetzmacher
# Copyright (C) 2019-2020 Daniel Ring
from ..scraper import ParserScraper


class KeenSpot(ParserScraper):
    multipleImagesPerStrip = True
    imageSearch = (
        '//img[contains(@src, "/comics/")]',
        # Gene Catlow Alternate
        '//img[contains(@src, "/altcomics/")]',
        # Shockwave Darkside
        '//img[contains(@src, "/comics2D/")]',
        '//img[contains(@src, "com/shockwave")]',
        # Sore Thumbs
        '//img[contains(@src, "com/st2")]',
        # Wayward Sons
        '//img[contains(@src, "com/2")]',
    )
    prevSearch = (
        '//link[@rel="prev"]',
        '//a[@rel="prev"]',
        # Exposure
        '//a[img[@id="exp29"]]',
        # Hero By Night
        '//area[contains(@coords, ",-7,")]',
        # Katrina
        '//a[img[@id="katc7"]]',
        # No Room For Magic, Everyone Loves Adis, Wisdom Of Moo
        '//a[text()="Previous comic"]',
        # Supernovas
        '//a[img[@id="p_top_nav"]]',
    )
    help = 'Index format: yyyymmdd'

    def __init__(self, name, sub, last=None, adult=False, path='d/%s.html'):
        super(KeenSpot, self).__init__('KeenSpot/' + name)
        self.url = 'http://%s.keenspot.com/' % sub
        self.stripUrl = self.url + path

        if last:
            self.url = self.stripUrl % last
            self.endOfLife = True

        if adult:
            self.adult = adult

    @classmethod
    def getmodules(cls):
        return (
            # Not on frontpage...
            cls('Buzzboy', 'buzzboy'),
            cls('EveryoneLovesAdis', 'adis'),
            cls('GeneCatlowAlternate', 'genecatlow', last='20170302',
                adult=True, path='altd/%s.html'),

            # do not edit anything below since these entries are generated from
            # scripts/update_plugins.sh
            # START AUTOUPDATE
            cls('27TwentySeven', 'twenty-seven'),
            cls('Avengelyne', 'avengelyne'),
            cls('BanzaiGirl', 'banzaigirl'),
            cls('Barker', 'barkercomic'),
            cls('ChoppingBlock', 'choppingblock'),
            cls('ClichFlamb', 'clicheflambe'),
            cls('CountYourSheep', 'countyoursheep'),
            cls('CrowScare', 'crowscare', last='20111031'),
            cls('Dreamless', 'dreamless', last='20100726'),
            cls('EverythingJake', 'everythingjake'),
            cls('Exposure', 'exposure'),
            cls('FallOutToyWorks', 'fallouttoyworks'),
            cls('FriarAndBrimstone', 'friarandbrimstone'),
            cls('GeneCatlow', 'genecatlow', last='20170412'),
            cls('GodMode', 'godmode'),
            cls('GreenWake', 'greenwake'),
            cls('HeadTrip', 'headtrip'),
            cls('HeroByNight', 'herobynight'),
            cls('HoaxHunters', 'hoaxhunters'),
            cls('InHere', 'inhere'),
            cls('JadeWarriors', 'jadewarriors'),
            cls('Katrina', 'katrina'),
            cls('LutherStrode', 'lutherstrode'),
            cls('MakeshiftMiracle', 'makeshiftmiracle'),
            cls('Marksmen', 'marksmen'),
            cls('MarryMe', 'marryme'),
            cls('MedusasDaughter', 'medusasdaughter'),
            cls('MonsterMassacre', 'monstermassacre'),
            cls('MysticRevolution', 'mysticrevolution', path='?cid=%s'),
            cls('NoPinkPonies', 'nopinkponies'),
            cls('NoRoomForMagic', 'noroomformagic'),
            cls('OutThere', 'outthere'),
            cls('Porcelain', 'porcelain'),
            cls('ProjectionEdge', 'newshounds'),
            cls('PunchAnPie', 'punchanpie', path='daily/%s.html'),
            cls('QUILTBAG', 'quiltbag'),
            cls('RedSpike', 'redspike'),
            cls('RumbleFall', 'rumblefall'),
            cls('SamuraisBlood', 'samuraisblood'),
            cls('Sharky', 'sharky'),
            cls('ShockwaveDarkside', 'shockwave', path='2d/%s.html'),
            cls('SomethingHappens', 'somethinghappens'),
            cls('SoreThumbs', 'sorethumbs'),
            cls('Striptease', 'striptease'),
            cls('Supernovas', 'supernovas'),
            cls('Superosity', 'superosity'),
            cls('TheFirstDaughter', 'thefirstdaughter'),
            cls('TheHopeVirus', 'hopevirus'),
            cls('TheHuntersOfSalamanstra', 'salamanstra'),
            cls('TheLounge', 'thelounge'),
            cls('TheVault', 'thevault'),
            cls('WaywardSons', 'waywardsons'),
            cls('WeirdingWillows', 'weirdingwillows'),
            cls('WICKEDPOWERED', 'wickedpowered'),
            cls('WisdomOfMoo', 'wisdomofmoo'),
            # END AUTOUPDATE
        )

    def shouldSkipUrl(self, url, data):
        return url in (
            'http://sorethumbs.keenspot.com/d/20160117.html'
        )



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: © 2004 Tristan Seligmann and Jonathan Jacobs
# SPDX-FileCopyrightText: © 2012 Bastian Kleineidam
# SPDX-FileCopyrightText: © 2015 Tobias Gruetzmacher
from __future__ import annotations

import html
import logging
import os
import re
import subprocess
import sys
import time
import traceback
from urllib import parse

import lxml

from . import AppName, http
from .configuration import App, SupportUrl

logger = logging.getLogger(__name__)

# Maximum content size for HTML pages
MaxContentBytes = 1024 * 1024 * 3  # 3 MB

# The character set to encode non-ASCII characters in a URL. See also
# http://tools.ietf.org/html/rfc2396#section-2.1
# Note that the encoding is not really specified, but most browsers
# encode in UTF-8 when no encoding is specified by the HTTP headers,
# else they use the page encoding for followed link. See als
# http://code.google.com/p/browsersec/wiki/Part1#Unicode_in_URLs
UrlEncoding = "utf-8"


def get_system_uid():
    """Get a (probably) unique ID to identify a system.
    Used to differentiate votes.
    """
    try:
        if os.name == 'nt':
            return get_nt_system_uid()
        if sys.platform == 'darwin':
            return get_osx_system_uid()
    except Exception:
        return get_mac_uid()
    else:
        return get_mac_uid()


def get_nt_system_uid():
    r"""Get the MachineGuid from
    HKEY_LOCAL_MACHINE\Software\Microsoft\Cryptography\MachineGuid
    """
    import winreg
    lm = winreg.ConnectRegistry(None, winreg.HKEY_LOCAL_MACHINE)
    try:
        key = winreg.OpenKey(lm, r"Software\Microsoft\Cryptography")
        try:
            return winreg.QueryValueEx(key, "MachineGuid")[0]
        finally:
            key.Close()
    finally:
        lm.Close()


def get_osx_system_uid():
    """Get the OSX system ID.
    $ system_profiler |grep "r (system)"
    Serial Number (system): C24E1322XYZ
    """
    res = backtick(["system_profile"]).splitlines()
    for line in res:
        if "r (system)" in line:
            return line.split(':', 1)[1].strip()
    raise ValueError("Could not find system number in %r" % res)


def get_mac_uid():
    """Get the MAC address of the system."""
    import uuid
    return "%d" % uuid.getnode()


def backtick(cmd, encoding='utf-8'):
    """Return decoded output from command."""
    data = subprocess.Popen(cmd, stdout=subprocess.PIPE).communicate()[0]
    return data.decode(encoding)


def unicode_safe(text, encoding=UrlEncoding, errors='ignore'):
    """Decode text to Unicode if not already done."""
    if isinstance(text, str):
        return text
    return text.decode(encoding, errors)


def tagre(tag, attribute, value, quote='"', before="", after=""):
    """Return a regular expression matching the given HTML tag, attribute
    and value. It matches the tag and attribute names case insensitive,
    and skips arbitrary whitespace and leading HTML attributes. The "<>" at
    the start and end of the HTML tag is also matched.
    @param tag: the tag name
    @ptype tag: string
    @param attribute: the attribute name
    @ptype attribute: string
    @param value: the attribute value
    @ptype value: string
    @param quote: the attribute quote (default ")
    @ptype quote: string
    @param after: match after attribute value but before end
    @ptype after: string

    @return: the generated regular expression suitable for re.compile()
    @rtype: string
    """
    if before:
        prefix = r"[^>]*%s[^>]*\s+" % before
    else:
        prefix = r"(?:[^>]*\s+)?"
    attrs = {
        'tag': case_insensitive_re(tag),
        'attribute': case_insensitive_re(attribute),
        'value': value,
        'quote': quote,
        'prefix': prefix,
        'after': after,
    }
    return (r'<\s*%(tag)s\s+%(prefix)s' +
            r'%(attribute)s\s*=\s*%(quote)s%(value)s%(quote)' +
            r's[^>]*%(after)s[^>]*>') % attrs


def case_insensitive_re(name):
    """Reformat the given name to a case insensitive regular expression string
    without using re.IGNORECASE. This way selective strings can be made case
    insensitive.
    @param name: the name to make case insensitive
    @ptype name: string
    @return: the case insensitive regex
    @rtype: string
    """
    return "".join("[%s%s]" % (c.lower(), c.upper()) for c in name)


def get_page(url, session, **kwargs):
    """Get text content of given URL."""
    http.check_robotstxt(url, session)
    # read page data
    page = urlopen(url, session, max_content_bytes=MaxContentBytes, **kwargs)
    logger.trace("Got page content %r", page.content)
    return page


def makeSequence(item):
    """If item is already a list or tuple, return it.
    Else return a tuple with item as single element."""
    if isinstance(item, (list, tuple)):
        return item
    return (item,)


def prettyMatcherList(things):
    """Try to construct a nicely-formatted string for a list of matcher
    objects. Those may be compiled regular expressions or strings..."""
    norm = []
    for x in makeSequence(things):
        if hasattr(x, 'pattern'):
            norm.append(x.pattern)
        else:
            norm.append(x)
    return "('%s')" % "', '".join(norm)


def normaliseURL(url):
    """Normalising
    - strips and leading or trailing whitespace,
    - replaces HTML entities and character references,
    - removes any leading empty segments to avoid breaking urllib2.
    """
    url = unicode_safe(url).strip()
    # XXX: brutal hack
    url = html.unescape(url)

    pu = list(parse.urlparse(url))
    segments = pu[2].split('/')
    while segments and segments[0] in ('', '..'):
        del segments[0]
    pu[2] = '/' + '/'.join(segments)
    # remove leading '&' from query
    if pu[4].startswith('&'):
        pu[4] = pu[4][1:]
    # remove anchor
    pu[5] = ""
    return parse.urlunparse(pu)


def urlopen(url, session, referrer=None, max_content_bytes=None,
            allow_errors=(), **kwargs):
    """Open an URL and return the response object."""
    logger.debug('Open URL %r', url)
    if 'headers' not in kwargs:
        kwargs['headers'] = {}
    if referrer:
        kwargs['headers']['Referer'] = referrer
    logger.trace('Sending headers %r', kwargs['headers'])
    logger.debug('Sending cookies %r', session.cookies)
    if 'data' not in kwargs:
        method = 'GET'
    else:
        method = 'POST'
        logger.trace('Sending POST data %r', kwargs['data'])
    req = session.request(method, url, **kwargs)
    logger.debug('Response cookies: %r', req.cookies)
    check_content_size(url, req.headers, max_content_bytes)
    if req.status_code not in allow_errors:
        req.raise_for_status()
    return req


def check_content_size(url, headers, max_content_bytes):
    """Check that content length in URL response headers do not exceed the
    given maximum bytes.
    """
    if not max_content_bytes:
        return
    if 'content-length' in headers:
        size = int(headers['content-length'])
        if size > max_content_bytes:
            raise IOError(
                'URL content of %s with %d bytes exceeds %d bytes.' %
                (url, size, max_content_bytes))


def splitpath(path):
    """Split a path in its components."""
    c = []
    head, tail = os.path.split(path)
    while tail:
        c.insert(0, tail)
        head, tail = os.path.split(head)
    return c


def getRelativePath(basepath, path):
    """Get a path that is relative to the given base path."""
    basepath = splitpath(os.path.abspath(basepath))
    path = splitpath(os.path.abspath(path))
    afterCommon = False
    for c in basepath:
        if afterCommon or path[0] != c:
            path.insert(0, os.path.pardir)
            afterCommon = True
        else:
            del path[0]
    return os.path.join(*path)


def getQueryParams(url):
    """Get URL query parameters."""
    query = parse.urlsplit(url).query
    logger.debug('Extracting query parameters from %r (%r)...', url, query)
    return parse.parse_qs(query)


def internal_error(out=sys.stderr, etype=None, evalue=None, tb=None):
    """Print internal error message (output defaults to stderr)."""
    print(os.linesep, file=out)
    print(f"""********** Oops, I did it again. *************

You have found an internal error in {AppName}. Please write a bug report
at {SupportUrl} and include at least the information below:

Not disclosing some of the information below due to privacy reasons is ok.
I will try to help you nonetheless, but you have to give me something
I can work with ;) .
""", file=out)
    if etype is None:
        etype = sys.exc_info()[0]
    if evalue is None:
        evalue = sys.exc_info()[1]
    print(etype, evalue, file=out)
    if tb is None:
        tb = sys.exc_info()[2]
    traceback.print_exception(etype, evalue, tb, None, out)
    print_app_info(out=out)
    print_proxy_info(out=out)
    print_locale_info(out=out)
    print(os.linesep,
          f"******** {AppName} internal error, over and out ********",
          file=out)


def print_env_info(key, out=sys.stderr):
    """If given environment key is defined, print it out."""
    value = os.getenv(key)
    if value is not None:
        print(key, "=", repr(value), file=out)


def print_proxy_info(out=sys.stderr):
    """Print proxy info."""
    print_env_info("http_proxy", out=out)


def print_locale_info(out=sys.stderr):
    """Print locale info."""
    for key in ("LANGUAGE", "LC_ALL", "LC_CTYPE", "LANG"):
        print_env_info(key, out=out)


def print_app_info(out=sys.stderr):
    """Print system and application info (output defaults to stderr)."""
    print("System info:", file=out)
    print(App, file=out)
    print("Python %(version)s on %(platform)s" %
          {"version": sys.version, "platform": sys.platform}, file=out)
    print("libxml2 version: %i.%i.%i" % lxml.etree.LIBXML_VERSION, file=out)
    stime = strtime(time.time())
    print("Local time:", stime, file=out)
    print("sys.argv", sys.argv, file=out)


def strtime(t):
    """Return ISO 8601 formatted time."""
    return (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(t)) +
            strtimezone())


def strtimezone():
    """Return timezone info, %z on some platforms, but not supported on all.
    """
    if time.daylight:
        zone = time.altzone
    else:
        zone = time.timezone
    return "%+04d" % (-zone // 3600)


def rfc822date(indate):
    """Format date in rfc822 format."""
    return time.strftime("%a, %d %b %Y %H:%M:%S GMT", time.gmtime(indate))


def unquote(text):
    """Replace all percent-encoded entities in text."""
    while '%' in text:
        newtext = parse.unquote(text)
        if newtext == text:
            break
        text = newtext
    return text


def getFilename(name):
    """Get a filename from given name without dangerous or incompatible
    characters."""
    # first replace all illegal chars
    name = re.sub(r"[^0-9a-zA-Z_\-\.]", "_", name)
    # then remove double dots and underscores
    while ".." in name:
        name = name.replace('..', '.')
    while "__" in name:
        name = name.replace('__', '_')
    # remove a leading dot or minus
    if name.startswith((".", "-")):
        name = name[1:]
    return name


def getExistingFile(name, max_suffix=1000):
    """Add filename suffix until file exists
    @return: filename if file is found
    @raise: ValueError if maximum suffix number is reached while searching
    """
    num = 1
    stem, ext = os.path.splitext(name)
    filename = name
    while not os.path.exists(filename):
        suffix = "-%d" % num
        filename = stem + suffix + ext
        num += 1
        if num >= max_suffix:
            raise ValueError("No file %r found" % name)
    return filename


def getNonexistingFile(name):
    """Add filename suffix until file not exists
    @return: filename
    """
    num = 1
    stem, ext = os.path.splitext(name)
    filename = name
    while os.path.exists(filename):
        suffix = "-%d" % num
        filename = stem + suffix + ext
        num += 1
    return filename


def strlimit(s, length=72):
    """If the length of the string exceeds the given limit, it will be cut
    off and three dots will be appended.

    @param s: the string to limit
    @type s: string
    @param length: maximum length
    @type length: non-negative integer
    @return: limited string, at most length+3 characters long
    """
    assert length >= 0, "length limit must be a non-negative integer"
    if not s or len(s) <= length:
        return s
    if length == 0:
        return ""
    return "%s..." % s[:length]


def uniq(input):
    """Remove duplicates from a list while preserving the list order"""
    output = []
    for item in input:
        if item not in output:
            output.append(item)
    return output


def urlpathsplit(url: str) -> list[str]:
    """Split the path of an URL into components, removing empty leading and
    trailing segments. This makes URL handling more robust against added or
    removed slashes.
    """
    parts = parse.urlsplit(url).path.split('/')
    while parts and not parts[-1]:
        del parts[-1]
    while parts and not parts[0]:
        del parts[0]
    return parts



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: © 2004 Tristan Seligmann and Jonathan Jacobs
# SPDX-FileCopyrightText: © 2012 Bastian Kleineidam
# SPDX-FileCopyrightText: © 2015 Tobias Gruetzmacher
from __future__ import annotations

import html
import logging
import os
import pathlib
import re
import warnings
from typing import Collection, Dict, List, Optional, Pattern, Sequence, Union
from urllib.parse import urljoin

import lxml
from lxml.html.defs import link_attrs as lxml_link_attrs

try:
    import pycountry
except ImportError:
    pycountry = None

from . import configuration, http, languages, loader, util, xml
from .comic import ComicStrip
from .events import getHandler

logger = logging.getLogger(__name__)
ARCHIVE_ORG_URL = re.compile(r'https?://web\.archive\.org/web/[^/]*/')
html_link_attrs = lxml_link_attrs - {'usemap'}


if lxml.etree.LIBXML_VERSION < (2, 9, 3):
    warnings.warn(  # noqa: B028 - false positive
        'Your libxml2 is very old (< 2.9.3), some dosage modules might missbehave')


class GeoblockedException(IOError):
    def __init__(self):
        super().__init__('It seems your current location is geo-blocked.')


class Scraper:
    '''Base class for all comic scraper, but without a specific scrape
    implementation.'''

    # The URL for the comic strip
    url: Optional[str] = None

    # A string that is interpolated with the strip index to yield the URL for a
    # particular strip.
    stripUrl: Optional[str] = None

    # Stop search for previous URLs at this URL
    firstStripUrl: Optional[str] = None

    # if more than one image per URL is expected
    multipleImagesPerStrip: bool = False

    # set to True if this comic contains adult content
    adult: bool = False

    # set to True if this comic will not get updated anymore
    endOfLife: bool = False

    # langauge of the comic (two-letter ISO 639-1 code)
    lang: str = 'en'

    # an expression that will locate the URL for the previous strip in a page
    # this can also be a list or tuple
    prevSearch: Optional[Union[Sequence[Union[str, Pattern]], str, Pattern]] = None

    # an expression that will locate the strip image URLs strip in a page
    # this can also be a list or tuple
    imageSearch: Optional[Union[Sequence[Union[str, Pattern]], str, Pattern]] = None

    # an expression to store a text together with the image
    # sometimes comic strips have additional text info for each comic
    textSearch: Optional[Union[Sequence[Union[str, Pattern]], str, Pattern]] = None

    # Is the additional text required or optional?  When it is required (the
    # default), you see an error message whenever a comic page is encountered
    # that does not have the text
    textOptional: bool = False

    # usually the index format help
    help: str = ''

    # Specifing a list of HTTP error codes which should be handled as a
    # successful request.  This is a workaround for some comics which return
    # regular pages with strange HTTP codes. By default, all HTTP errors raise
    # exceptions.
    allow_errors: Sequence[int] = ()

    # HTTP session for configuration & cookies
    session: http.Session = http.default_session

    @classmethod
    def getmodules(cls) -> Collection[Scraper]:
        if cls.url is None:
            return ()
        name = cls.__name__
        if hasattr(cls, 'name'):
            name = cls.name
        return (cls(name),)

    @property
    def indexes(self):
        return self._indexes

    @indexes.setter
    def indexes(self, val):
        if val:
            self._indexes = tuple(sorted(val))

    def __init__(self, name: str) -> None:
        """Initialize internal variables."""
        self.name = name
        self.urls: set[str] = set()
        self._indexes = ()
        self.skippedUrls: set[str] = set()
        self.hitFirstStripUrl = False

    def __hash__(self) -> int:
        """Get hash value from name and index list."""
        return hash((self.name, self.indexes))

    def shouldSkipUrl(self, url: str, data) -> bool:
        """Determine if search for images in given URL should be skipped."""
        return False

    def getComicStrip(self, url, data) -> ComicStrip:
        """Get comic strip downloader for given URL and data."""
        urls = self.extract_image_urls(url, data)
        # map modifier function on image URLs
        urls = [self.imageUrlModifier(x, data) for x in urls]
        # remove duplicate URLs
        urls = util.uniq(urls)
        if len(urls) > 1 and not self.multipleImagesPerStrip:
            logger.warning("Found %d images instead of 1 at %s with expressions %r",
                len(urls), url, util.prettyMatcherList(self.imageSearch))
            image = urls[0]
            logger.warning("Choosing image %r", image)
            urls = (image,)
        elif not urls:
            logger.warning("Found no images at %r with expressions %r", url,
                util.prettyMatcherList(self.imageSearch))
        if self.textSearch:
            text = self.fetchText(url, data, self.textSearch,
                optional=self.textOptional)
        else:
            text = None
        return ComicStrip(self, url, urls, text=text)

    def getStrips(self, maxstrips=None):
        """Get comic strips."""
        if maxstrips:
            word = u"strip" if maxstrips == 1 else "strips"
            msg = u'Retrieving %d %s' % (maxstrips, word)
        else:
            msg = u'Retrieving all strips'
        if self.indexes:
            if len(self.indexes) == 1:
                msg += u" for index %s" % self.indexes[0]
            else:
                msg += u" for indexes %s" % self.indexes
            # Always call starter() since it might initialize cookies.
            # See for example Oglaf comic.
            self.starter()
            urls = [self.getIndexStripUrl(index) for index in self.indexes]
        else:
            urls = [self.starter()]
        if self.adult:
            msg += u" (including adult content)"
        logger.info(msg)
        for url in urls:
            for strip in self.getStripsFor(url, maxstrips):
                yield strip

    def getStripsFor(self, url, maxstrips):
        """Get comic strips for an URL. If maxstrips is a positive number, stop after
        retrieving the given number of strips."""
        self.hitFirstStripUrl = False
        seen_urls = set()
        while url:
            logger.moreinfo('Get strip URL %r', url)
            data = self.getPage(url)
            if self.shouldSkipUrl(url, data):
                logger.info('Skipping URL %r', url)
                self.skippedUrls.add(url)
            else:
                try:
                    yield self.getComicStrip(url, data)
                except ValueError as msg:
                    logger.exception(msg)  # noqa: LOG010
            if self.isfirststrip(url):
                logger.debug("Stop at first URL %r", url)
                self.hitFirstStripUrl = True
                break
            if maxstrips is not None:
                maxstrips -= 1
                if maxstrips <= 0:
                    break
            prevUrl = self.getPrevUrl(url, data)
            seen_urls.add(url)
            if prevUrl in seen_urls:
                # avoid recursive URL loops
                logger.warning("Already seen previous URL %r", prevUrl)
                break
            url = prevUrl

    def isfirststrip(self, url: str) -> bool:
        """Check if the specified URL is the first strip of a comic. This is
        specially for comics taken from archive.org, since the base URL of
        archive.org changes whenever pages are taken from a different
        snapshot."""
        if not self.firstStripUrl:
            return False
        firsturl = ARCHIVE_ORG_URL.sub('', self.firstStripUrl)
        currenturl = ARCHIVE_ORG_URL.sub('', url)
        return firsturl == currenturl

    def getPrevUrl(self, url: str, data) -> str | None:
        """Find previous URL."""
        prevUrl = None
        if self.prevSearch:
            try:
                prevUrl = self.fetchUrl(url, data, self.prevSearch)
            except ValueError as msg:
                # assume there is no previous URL, but print a warning
                logger.warning("%s Assuming no previous comic strips exist.", msg)
            else:
                prevUrl = self.link_modifier(url, prevUrl)
                logger.debug("Found previous URL %r", prevUrl)
                getHandler().comicPageLink(self, url, prevUrl)
        return prevUrl

    def getIndexStripUrl(self, index: str) -> str:
        """Get comic strip URL from index."""
        if not self.stripUrl:
            raise ValueError("Getting by index is not supported!")
        return self.stripUrl % index

    def starter(self) -> str | None:
        """Get starter URL from where to scrape comic strips."""
        return self.url

    def namer(self, image_url: str, page_url: str) -> str:
        """Return filename for given image and page URL."""
        return util.urlpathsplit(image_url)[-1]

    def link_modifier(self, fromurl: str, tourl: str) -> str:
        """Optional modification of parsed link (previous/back/latest) URLs.
        Useful if there are domain redirects. The default implementation does
        not modify the URL.
        """
        return tourl

    def imageUrlModifier(self, image_url: str, data) -> str:
        """Optional modification of parsed image URLs. Useful if the URL
        needs to be fixed before usage. The default implementation does
        not modify the URL. The given data is the URL page data.
        """
        return image_url

    def vote(self) -> None:
        """Cast a public vote for this comic."""
        uid = util.get_system_uid()
        data = {"name": self.name.replace('/', '_'), "uid": uid}
        response = self.session.post(configuration.VoteUrl, data=data)
        response.raise_for_status()

    def get_download_dir(self, basepath: str) -> str:
        """Try to find the corect download directory, ignoring case
        differences."""
        path = basepath
        for part in self.name.split('/'):
            done = False
            if (os.path.isdir(path) and
               not os.path.isdir(os.path.join(path, part))):
                for entry in os.listdir(path):
                    if (entry.lower() == part.lower() and
                       os.path.isdir(os.path.join(path, entry))):
                        path = os.path.join(path, entry)
                        done = True
                        break
            if not done:
                path = os.path.join(path, part)
        return path

    def getCompleteFile(self, basepath: str) -> str:
        """Get filename indicating all comics are downloaded."""
        dirname = self.get_download_dir(basepath)
        return os.path.join(dirname, "complete.txt")

    def isComplete(self, basepath: str) -> bool:
        """Check if all comics are downloaded."""
        return os.path.isfile(self.getCompleteFile(basepath))

    def setComplete(self, basepath: str) -> None:
        """Set complete flag for this comic, ie. all comics are downloaded."""
        if self.endOfLife:
            filename = self.getCompleteFile(basepath)
            if not os.path.exists(filename):
                with open(filename, 'w') as f:
                    f.write('All comics should be downloaded here.')

    def getPage(self, url):
        """
        Fetch a page and return the opaque repesentation for the data parameter
        of fetchUrls and fetchText.

        Implementation notes: While this base class does not restrict how the
        returned data is structured, subclasses (specific scrapers) should
        specify how this data works, since the stracture is passed into
        different methods which can be defined by comic modules and these
        methods should be able to use the data if they so desire... (Affected
        methods: shouldSkipUrl, imageUrlModifier)
        """
        return util.get_page(url, self.session, allow_errors=self.allow_errors)

    def extract_image_urls(self, url, data):
        """
        Extract image URLs from page data using the classes imageSearch attribute.
        """
        return self.fetchUrls(url, data, self.imageSearch)

    def fetchUrls(self, url, data, urlsearch):
        raise ValueError("No implementation for fetchUrls!")

    def fetchUrl(self, url, data, urlsearch):
        return self.fetchUrls(url, data, urlsearch)[0]

    def fetchText(self, url, data, textsearch, optional):
        raise ValueError("No implementation for fetchText!")

    def getDisabledReasons(self):
        """
        Get a dict of reasons why this comic module is disabled. The key is a
        short (unique) identifier, the value is a string explaining why the
        module is deactivated. If the module is not disabled, just return an
        empty dict.
        """
        return {}

    def language(self):
        """
        Return language of the comic as a human-readable language name instead
        of a 2-character ISO639-1 code.
        """
        if pycountry is None:
            if self.lang in languages.Languages:
                return languages.Languages[self.lang]
        else:
            try:
                return pycountry.languages.get(alpha_2=self.lang).name
            except KeyError:
                try:
                    return pycountry.languages.get(alpha2=self.lang).name
                except KeyError:
                    pass
        return 'Unknown (%s)' % self.lang

    def geoblocked(self):
        """Helper method to indicate that the user is most probably geo-blocked."""
        raise GeoblockedException()


class BasicScraper(Scraper):
    """
    Scraper base class that matches regular expressions against HTML pages.

    Subclasses of this scraper should use compiled regular expressions as
    values for prevSearch, imageSearch and textSearch.

    Implementation note: The return value of getPage is a tuple: the first
    element is the raw HTML page text, the second element is the base URL (if
    any).
    """

    BASE_SEARCH = re.compile(util.tagre("base", "href", '([^"]*)'))

    def getPage(self, url):
        content = super().getPage(url).text
        # determine base URL
        baseUrl = None
        match = self.BASE_SEARCH.search(content)
        if match:
            baseUrl = match.group(1)
        else:
            baseUrl = url
        return (content, baseUrl)

    def fetchUrls(self, url, data, urlSearch):
        """Search all entries for given URL pattern(s) in a HTML page."""
        searchUrls = []
        searches = util.makeSequence(urlSearch)
        for search in searches:
            for match in search.finditer(data[0]):
                searchUrl = match.group(1)
                if not searchUrl:
                    raise ValueError("Pattern %s matched empty URL at %s." %
                                     (search.pattern, url))
                logger.debug('matched URL %r with pattern %r', searchUrl, search.pattern)
                searchUrls.append(util.normaliseURL(urljoin(data[1], searchUrl)))
            if searchUrls:
                # do not search other links if one pattern matched
                break
        if not searchUrls:
            patterns = [x.pattern for x in searches]
            raise ValueError("Patterns %s not found at URL %s." %
                             (patterns, url))
        return searchUrls

    def fetchText(self, url, data, textSearch, optional):
        """Search text entry for given text pattern in a HTML page."""
        if textSearch:
            match = textSearch.search(data[0])
            if match:
                text = match.group(1)
                logger.debug('matched text %r with pattern %r', text, textSearch.pattern)
                return html.unescape(text).strip()
            if optional:
                return None
            else:
                raise ValueError("Pattern %s not found at URL %s." %
                                 (textSearch.pattern, url))
        else:
            return None


class ParserScraper(Scraper):
    """
    Scraper base class that uses a HTML parser and XPath expressions.

    All links are resolved before XPath searches are applied, so all URLs are
    absolute!

    Subclasses of this class should use XPath expressions as values for
    prevSearch, imageSearch and textSearch. When the XPath directly selects an
    attribute, it is used as the output.

    All those searches try to do something intelligent when they match a
    complete HTML Element: prevSearch and imageSearch try to find a "link
    attribute" and use that as URL. textSearch strips all tags from the content
    of the HTML element and returns that.
    """

    # Taken directly from LXML
    XML_DECL = re.compile(
        r'^(<\?xml[^>]+)\s+encoding\s*=\s*["\'][^"\']*["\'](\s*\?>|)', re.U)

    def getPage(self, url):
        page = super().getPage(url)
        if page.encoding:
            # Requests figured out the encoding, so we can deliver Unicode to
            # LXML. Unfortunatly, LXML feels betrayed if there is still an XML
            # declaration with (probably wrong!) encoding at the top of the
            # document. Web browsers ignore such if the encoding was specified
            # in the HTTP header and so do we.
            text = self.XML_DECL.sub('\1\2', page.text, count=1)
            tree = self._parse_page(text)
        else:
            tree = self._parse_page(page.content)
        tree.make_links_absolute(url)
        return tree

    def _parse_page(self, data):
        return lxml.html.document_fromstring(data)

    def fetchUrls(self, url, data, urlSearch):
        """Search all entries for given XPath in a HTML page."""
        searchUrls = []
        for match, search in self._matchPattern(data, urlSearch):
            searchUrl = None
            try:
                for attrib in html_link_attrs:
                    if attrib in match.attrib:
                        searchUrl = match.get(attrib)
            except AttributeError:
                searchUrl = str(match)
            logger.debug('Matched URL %r with pattern %r', searchUrl, search)
            if searchUrl is not None:
                searchUrls.append(searchUrl)

        if not searchUrls:
            raise ValueError("XPath %s not found at URL %s." %
                             (urlSearch, url))
        return searchUrls

    def fetchText(self, url, data, textSearch, optional):
        """Search text entry for given text XPath in a HTML page."""
        if not textSearch:
            return None
        text = []
        for match, search in self._matchPattern(data, textSearch):
            try:
                text.append(match.text_content())
            except AttributeError:
                text.append(match)
            logger.debug('Matched text %r with XPath %r', text, search)
        text = u' '.join(text)
        if text.strip() == '':
            if optional:
                return None
            else:
                raise ValueError("XPath %s did not match anything at URL %s." %
                                 (textSearch, url))
        return text.strip()

    def _matchPattern(self, data, patterns):
        patterns = util.makeSequence(patterns)
        for search in patterns:
            matched = False
            for match in self.match(data, search):
                matched = True
                yield match, search

            if matched and not self.multipleImagesPerStrip:
                # do not search other links if one pattern matched
                break

    def match(self, data, pattern):
        """Match an XPath pattern against a page."""
        return data.xpath(pattern, namespaces=xml.NS)


# Legacy aliases
_BasicScraper = BasicScraper
_ParserScraper = ParserScraper


class Cache:
    """Cache for comic scraper objects. The cache is initialized on first use.
    This is cached, since iterating & loading a complete package might be quite
    slow.
    """
    def __init__(self) -> None:
        self.data: List[Scraper] = []
        self.userdirs: set[pathlib.Path] = set()

    def find(self, comic: str) -> Scraper:
        """Find a comic scraper object based on its name. This prefers a
        perfect match, but falls back to a substring match, if that is unique.
        Otharwise a ValueError is thrown.
        """
        if not comic:
            raise ValueError("empty comic name")
        candidates = []
        cname = comic.lower()
        for scraper in self.all(include_removed=True):
            lname = scraper.name.lower()
            if lname == cname:
                # perfect match
                return scraper
            elif cname in lname and scraper.url:
                candidates.append(scraper)
        if len(candidates) > 1:
            comics = ", ".join(x.name for x in candidates)
            raise ValueError('multiple comics found: %s' % comics)
        elif not candidates:
            raise ValueError('comic %r not found' % comic)
        return candidates[0]

    def load(self) -> None:
        logger.debug("Loading comic modules...")
        modules = 0
        classes = 0
        for module in loader.get_plugin_modules():
            modules += 1
            classes += self.addmodule(module)
        self.validate()
        logger.debug("... %d scrapers loaded from %d classes in %d modules.",
            len(self.data), classes, modules)

    def adddir(self, path: pathlib.Path) -> None:
        """Add an additional directory with python modules to the scraper list.
        These are handled as if the were part of the plugins package.
        """
        if path in self.userdirs:
            return
        if not self.data:
            self.load()
        modules = 0
        classes = 0
        logger.debug("Loading user scrapers from %r...", path)
        for module in loader.get_plugin_modules_from_dir(path):
            modules += 1
            classes += self.addmodule(module)
        self.validate()
        self.userdirs.add(path)
        if classes > 0:
            logger.debug("Added %d user classes from %d modules.",
                classes, modules)

    def addmodule(self, module) -> int:
        """Adds all valid plugin classes from the specified module to the cache.
        @return: number of classes added
        """
        classes = 0
        for plugin in loader.get_module_plugins(module, Scraper):
            classes += 1
            self.data.extend(plugin.getmodules())
        return classes

    def all(self, include_removed=False) -> list[Scraper]:
        """Return all comic scraper classes in the plugins directory.
        @return: list of Scraper classes
        @rtype: list of Scraper
        """
        if not self.data:
            self.load()
        if include_removed:
            return self.data
        else:
            return [x for x in self.data if x.url]

    def validate(self) -> None:
        """Check for duplicate scraper names."""
        d: Dict[str, Scraper] = {}
        for scraper in self.data:
            name = scraper.name.lower()
            if name in d:
                name1 = scraper.name
                name2 = d[name].name
                raise ValueError('duplicate scrapers %s and %s found' %
                                 (name1, name2))
            d[name] = scraper


scrapers = Cache()



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: © 2004 Tristan Seligmann and Jonathan Jacobs
# SPDX-FileCopyrightText: © 2012 Bastian Kleineidam
# SPDX-FileCopyrightText: © 2015 Tobias Gruetzmacher
"""
Script to get arcamax comics and save the info in a JSON file for further
processing.
"""

import scriptutil


class ArcamaxUpdater(scriptutil.ComicListUpdater):
    dup_templates = ("Creators/%s", "DrunkDuck/%s", "GoComics/%s",
                     "KeenSpot/%s", "ComicGenesis/%s")

    # names of comics to exclude
    excluded_comics = (
        # better source available
        "Dilbert",
        "HagarTheHorrible",
    )

    def handle_url(self, url):
        """Parse one search result page."""
        data = self.get_url(url)

        for comiclink in self.xpath(data, '//a[d:class("comic-icon")]'):
            path = comiclink.attrib['href']
            name = comiclink.attrib['title']

            self.add_comic(name, path.rsplit('/', 2)[1])

    def collect_results(self):
        """Parse all search result pages."""
        self.handle_url('http://www.arcamax.com/comics')

    def get_entry(self, name, entry):
        return u"cls('%s', '%s')," % (name, entry)


if __name__ == '__main__':
    ArcamaxUpdater(__file__).run()



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: © 2019 Tobias Gruetzmacher
# SPDX-FileCopyrightText: © 2019 Daniel Ring
"""
Script to get a list of Tapastic comics and save the info in a
JSON file for further processing.
"""
from urllib import parse

from scriptutil import ComicListUpdater


class TapasUpdater(ComicListUpdater):
    def collect_results(self):
        # Retrieve the first 10 top comics list pages
        url = 'https://tapas.io/comics?browse=ALL&sort_type=LIKE&pageNumber='
        count = 10

        data = [self.get_url(url + str(i), robot=False) for i in range(0, count)]
        for page in data:
            for comiclink in page.xpath('//a[@class="preferred title"]'):
                comicurl = comiclink.attrib['href']
                name = comiclink.text
                self.add_comic(name, comicurl)

    def get_entry(self, name, url):
        shortName = name.replace(' ', '').replace('\'', '')
        titleNum = int(parse.parse_qs(parse.urlsplit(url).query)['title_no'][0])
        url = url.rsplit('/', 1)[0].replace('/series/', '')
        return u"cls('%s', '%s', %d)," % (shortName, url, titleNum)


if __name__ == '__main__':
    TapasUpdater(__file__).run()



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: © 2004 Tristan Seligmann and Jonathan Jacobs
# SPDX-FileCopyrightText: © 2012 Bastian Kleineidam
# SPDX-FileCopyrightText: © 2015 Tobias Gruetzmacher
# SPDX-FileCopyrightText: © 2019 Daniel Ring
from re import compile
from urllib.parse import urljoin

from .. import util
from ..helpers import bounceStarter, joinPathPartsNamer
from ..scraper import ParserScraper, _BasicScraper, _ParserScraper
from ..util import tagre
from .common import WordPressNavi, WordPressScraper, WordPressWebcomic


class Rainburn(ParserScraper):
    url = 'https://rainburn.virmir.com/'
    stripUrl = url + 'comic.php?page=%s'
    firstStripUrl = stripUrl % '1'
    imageSearch = '//div[@class="comic"]//img'
    prevSearch = '//a[contains(@class, "prev")]'


class RalfTheDestroyer(WordPressScraper):
    url = 'http://ralfthedestroyer.com/'


class RayFox(WordPressNavi):
    url = 'https://www.rayfoxthecomic.com/'
    stripUrl = url + 'comic/%s/'
    firstStripUrl = stripUrl % 'not-a-super-hero/it-begins'

    def namer(self, image_url, page_url):
        filename = util.urlpathsplit(image_url)[-1].split('.', 1)[0]
        if filename == 'j':
            filename = 'RF_E3_P52'
        elif filename == '46' or filename == '55' or filename == '61':
            filename = 'RF_E3_P' + filename
        elif 'chapter-3-cover' in filename:
            filename = 'RF_E3_Cover'
        elif 'Cover2' in filename:
            filename = 'RF_E1_' + filename
        elif 'Volume-1-Cover' in filename:
            filename = filename.replace('Ray-Fox-Volume-1-', 'RF_E1_')
        elif filename[0] == '0':
            filename = 'RF_E1_P' + filename
        return filename


class RaynaOnTheRiver(WordPressScraper):
    url = 'http://www.catomix.com/rayna/'
    firstStripUrl = url + 'archives/comic/teaser-poster'


class RealLife(WordPressScraper):
    url = 'https://reallifecomics.com/'
    stripUrl = url + 'comic.php?comic=%s'
    firstStripUrl = stripUrl % 'title-1'
    help = 'Index format: monthname-dd-yyyy'

    def namer(self, image_url, page_url):
        # Fix inconsisntent filenames
        filename = util.urlpathsplit(image_url)[-1]
        if page_url.rsplit('=', 1)[-1] == 'may-27-2014':
            filename = filename.replace('20140219_3121', '20140527')
        filename = filename.replace('5-Finished', '20140623_3161')
        filename = filename.replace('520140722', '20140722')
        filename = filename.replace('520140724', '20140724')
        return filename

    def getPrevUrl(self, url, data):
        # "Parse" JavaScript
        prevtag = data.find_class('comic-nav-previous')
        if not prevtag:
            return None
        target = prevtag[0].get('onclick').split("'")[1]
        return urljoin(url, target)


class RealmOfAtland(_ParserScraper):
    url = 'https://web.archive.org/web/20201225151926/http://www.realmofatland.com/'
    stripUrl = url + '?p=%s'
    firstStripUrl = stripUrl % '1'
    prevSearch = '//a[@id="cg_back"]'
    imageSearch = '//p[@id="cg_img"]//img'
    endOfLife = True
    help = 'Index format: nnn'


class Recursion(_ParserScraper):
    url = 'https://recursioncomic.com/'
    stripUrl = url + '%s'
    firstStripUrl = stripUrl % '0001'
    imageSearch = '//div[@class="content"]//img'
    prevSearch = '//link[@rel="prev"]'

    def namer(self, image_url, page_url):
        # Fix inconsistent filenames
        filename = util.urlpathsplit(image_url)[-1]
        filename = filename.replace('0bf62e92-2c98-4fb2-8ed7-4584980beb17', 'page0005')
        filename = filename.replace('76112837-c5cd-4df7-8c53-3ed5c25194cf', 'page0003')
        filename = filename.replace('ed271080-6b1b-4d7a-8509-b2d8a15da805', 'page0002')
        filename = filename.replace('7b194ef7-ac77-4b5c-aed0-826901d13d04', 'page0001')
        return filename


class RedMeat(ParserScraper):
    url = 'https://www.redmeat.com/'
    imageSearch = '//article[d:class("comic")]//img'
    prevSearch = '//a[d:class("prev")]'
    namer = joinPathPartsNamer(imageparts=range(-2, 0))


class Requiem(WordPressScraper):
    baseUrl = 'http://requiem.spiderforest.com/'
    url = baseUrl + '?latest'
    stripUrl = baseUrl + 'comic/%s'
    firstStripUrl = stripUrl % '2004-06-07-3'


class Replay(ParserScraper):
    url = 'http://replaycomic.com/'
    stripUrl = url + 'comic/%s/'
    firstStripUrl = stripUrl % 'red-desert'
    imageSearch = '//div[@id="comic"]//img[@alt]'
    prevSearch = '//a[contains(@class, "comic-nav-previous")]'
    nextSearch = '//a[contains(@class, "comic-nav-next")]'

    def starter(self):
        # Retrieve archive page to identify chapters
        archivePage = self.getPage(self.url + 'archive')
        archive = self.match(archivePage, '//div[d:class("comic-archive-chapter-wrap")]')
        self.chapter = len(archive) - 1
        self.startOfChapter = []
        for archiveChapter in archive:
            self.startOfChapter.append(self.match(archiveChapter, './/a')[0].get('href'))
        return bounceStarter(self)

    def namer(self, imageUrl, pageUrl):
        # Name pages based on chapter, index, and post title
        name = util.urlpathsplit(pageUrl)[-1]
        page = util.urlpathsplit(imageUrl)[-1].rsplit('.', 1)[0]

        # Fix inconsistent page number formatting
        if page.isdigit() and len(page) > 2 and self.chapter == 1 and name != 'through-the-woods':
            page = page[:2] + '-' + page[2:]

        name = f'{self.chapter}-{page}-{name}'
        if pageUrl in self.startOfChapter:
            self.chapter -= 1
        return name


class RiversideExtras(WordPressWebcomic):
    url = 'https://riversidecomics.com/'


class RomanticallyApocalyptic(_ParserScraper):
    url = 'http://romanticallyapocalyptic.com/'
    stripUrl = url + '%s'
    firstStripUrl = stripUrl % '0'
    imageSearch = '//div[d:class("comicpanel")]/center//img'
    prevSearch = '//a[@accesskey="p"]'
    help = 'Index format: n'
    adult = True


class Roza(_ParserScraper):
    url = 'http://www.junglestudio.com/roza/index.php'
    stripUrl = url + '?date=%s'
    firstStripUrl = stripUrl % '2007-05-01'
    imageSearch = '//img[contains(@src, "pages/")]'
    prevSearch = '//a[img[contains(@src, "navtable_01.gif")]]'
    help = 'Index format: yyyy-mm-dd'


class Ruthe(_BasicScraper):
    url = 'http://ruthe.de/'
    stripUrl = url + 'cartoon/%s/datum/asc/'
    firstStripUrl = stripUrl % '1'
    lang = 'de'
    imageSearch = compile(tagre("img", "src", r'(/?cartoons/strip_\d+[^"]+)'))
    prevSearch = compile(tagre("a", "href", r'(/cartoon/\d+/datum/asc/)'))
    help = 'Index format: number'



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: © 2004 Tristan Seligmann and Jonathan Jacobs
# SPDX-FileCopyrightText: © 2012 Bastian Kleineidam
# SPDX-FileCopyrightText: © 2015 Tobias Gruetzmacher
from __future__ import annotations

import os
import re
from typing import Any, Dict, Optional, Tuple

import dosagelib

from . import http

UPDATE_URL = 'https://api.github.com/repos/webcomics/dosage/releases/latest'
EXTPRIO = {
    '.exe': 1 if os.name == 'nt' else 9,
    '.whl': 2,
    '.gz': 3,
}


def check_update() -> Optional[Tuple[str, Optional[str]]]:
    """Return the following values:
       throws exception - online version could not be determined
       None - user has newest version
       (version, url string) - update available
       (version, None) - current version is newer than online version
    """
    version, value = get_online_version()
    if version == dosagelib.__version__:
        # user has newest version
        return None
    if is_newer_version(version):
        # value is an URL linking to the update package
        return (version, value)
    # user is running a local or development version
    return (version, None)


def asset_key(asset: Dict[str, Any]) -> int:
    return EXTPRIO.get(os.path.splitext(asset['browser_download_url'])[1], 99)


def get_online_version() -> Tuple[str, Optional[str]]:
    """Download update info and parse it."""
    response = http.default_session.get(UPDATE_URL)
    response.raise_for_status()
    page = response.json()
    version = page['tag_name']

    url = None
    try:
        assets = sorted(page['assets'], key=asset_key)
        if len(assets) > 0:
            url = assets[0]['browser_download_url']
    except KeyError:
        pass
    return version, url


def version_nums(ver: str) -> Tuple[int, ...]:
    """Extract all numeric "sub-parts" of a version string. Not very exact, but
       works for our use case."""
    return tuple(int(s) for s in re.split(r'\D+', ver + '0'))


def is_newer_version(version) -> bool:
    """Check if given version is newer than current version."""
    return version_nums(version) > version_nums(dosagelib.__version__)



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: © 2004 Tristan Seligmann and Jonathan Jacobs
# SPDX-FileCopyrightText: © 2012 Bastian Kleineidam
# SPDX-FileCopyrightText: © 2015 Tobias Gruetzmacher
# SPDX-FileCopyrightText: © 2019 Daniel Ring
from re import IGNORECASE, compile, escape

from .. import util
from ..helpers import bounceStarter, joinPathPartsNamer
from ..scraper import ParserScraper, _BasicScraper, _ParserScraper
from ..util import tagre
from .common import (
    ComicControlScraper,
    WordPressNaviIn,
    WordPressScraper,
    WordPressWebcomic,
)


class WapsiSquare(WordPressNaviIn):
    url = 'http://wapsisquare.com/'
    firstStripUrl = url + 'comic/09092001/'

    def shouldSkipUrl(self, url, data):
        """Skip pages without images."""
        return self.match(data, '//iframe')  # videos


class WastedTalent(_ParserScraper):
    url = 'http://www.wastedtalent.ca/'
    stripUrl = url + 'comic/%s'
    firstStripUrl = stripUrl % 'anime-crack'
    imageSearch = '//div[d:class("comic_content")]/img'
    prevSearch = '//li[d:class("previous")]/a'
    multipleImagesPerStrip = True


class WebcomicName(_ParserScraper):
    url = 'https://webcomicname.com/'
    imageSearch = '//figure[d:class("tmblr-full")]//img'
    prevSearch = '//a[d:class("next")]'
    multipleImagesPerStrip = True


class Weregeek(ParserScraper):
    url = 'http://www.weregeek.com/'
    stripUrl = url + 'comic/%s/'
    firstStripUrl = stripUrl % 'comic-1'
    imageSearch = '//div[d:class("webcomic-media")]//img'
    prevSearch = '//a[d:class("previous-webcomic-link")]'


class WereIWolf(_ParserScraper):
    stripUrl = 'https://wolfwares.ca/comics/Were I wolf/strip2.php?name=%s&start=%s'
    url = stripUrl % ('4 Black and White - part 3', 'latest')
    firstStripUrl = stripUrl % ('1 Sirens', '0')
    imageSearch = '//img[contains(@src, "ROW")]'
    prevSearch = '//a[./img[contains(@src, "previous")]]'
    multipleImagesPerStrip = True
    endOfLife = True
    chapters = ('1 Sirens',
                '2 Black and White',
                '3 Black and White - Princess and Knight',
                '4 Black and White - part 3')

    def namer(self, image_url, page_url):
        # Prepend chapter number to image filename
        for chapter in self.chapters:
            if chapter in page_url:
                chapterNum = chapter[0]
        return chapterNum + '_' + util.urlpathsplit(image_url)[-1]

    def getPrevUrl(self, url, data):
        # Fix missing navigation links between chapters
        if url == self.stripUrl % (self.chapters[3], '0'):
            return self.stripUrl % (self.chapters[2], 'latest')
        if url == self.stripUrl % (self.chapters[2], '0'):
            return self.stripUrl % (self.chapters[1], 'latest')
        if url == self.stripUrl % (self.chapters[1], '0'):
            return self.stripUrl % (self.chapters[0], 'latest')
        return super(WereIWolf, self).getPrevUrl(url, data)

    def getIndexStripUrl(self, index):
        # Get comic strip URL from index
        index = index.split('-')
        return self.stripUrl % (index[0], index[1])


class WhiteNoise(WordPressWebcomic):
    url = 'http://whitenoisecomic.com/'
    stripUrl = url + 'comic/%s/'
    firstStripUrl = stripUrl % 'book-one'
    imageSearch = '//div[@id="comic"]//img'


class WhiteNoiseLee(ComicControlScraper):
    url = 'http://www.white-noise-comic.com/'
    stripUrl = url + 'comic/%s'
    firstStripUrl = stripUrl % '1-0'
    starter = bounceStarter
    namer = joinPathPartsNamer(pageparts=(-1,))


class Whomp(ComicControlScraper):
    url = 'http://www.whompcomic.com/'
    firstStripUrl = url + 'comic/06152010'
    textSearch = '//img[@id="cc-comic"]/@title'


class WhyTheLongFace(_BasicScraper):
    baseUrl = 'http://www.absurdnotions.org/'
    rurl = escape(baseUrl)
    url = baseUrl + 'wtlf200709.html'
    stripUrl = baseUrl + 'wtlf%s.html'
    firstStripUrl = stripUrl % '200306'
    imageSearch = compile(r'<img src="(%swtlf.+?|lf\d+.\w{1,4})"' % rurl,
                          IGNORECASE)
    multipleImagesPerStrip = True
    prevSearch = compile(r'HREF="(.+?)"><IMG SRC="nprev.gif" ')
    help = 'Index format: yyyymm'


class Widdershins(ComicControlScraper):
    url = 'https://widdershinscomic.com/'
    stripUrl = url + 'wdshn/%s'
    firstStripUrl = stripUrl % 'sleight-of-hand-cover'
    starter = bounceStarter
    namer = joinPathPartsNamer(pageparts=(-1,))


class Wigu(_ParserScraper):
    stripUrl = 'http://www.wigucomics.com/adventures/index.php?comic=%s'
    url = stripUrl % '-1'
    firstStripUrl = stripUrl % '1'
    imageSearch = '//div[@id="comic"]//img[contains(@src, "/comics/")]'
    prevSearch = '//a[@alt="go back"]'
    endOfLife = True
    help = 'Index format: n'


class WildeLife(ComicControlScraper):
    url = 'http://www.wildelifecomic.com/'
    stripUrl = url + 'comic/%s'
    firstStripUrl = stripUrl % '1'


class Wolfpac(WordPressScraper):
    url = 'https://wolfpac.ca/'
    firstStripUrl = url + 'archives/comic/wolfpac-title'


class Wonderella(_BasicScraper):
    url = 'http://nonadventures.com/'
    rurl = escape(url)
    stripUrl = url + '%s/'
    firstStripUrl = stripUrl % '2006/09/09/the-torment-of-a-thousand-yesterdays'
    imageSearch = compile(tagre("div", "id", r"comic", quote=r'["\']') +
                          r"\s*" +
                          tagre("img", "src", r'(%scomics/[^"]+)' % rurl))
    prevSearch = compile(tagre("a", "href", r'(%s\d+/\d+/\d+/[^"]+)' % rurl,
                               after="prev"))
    help = 'Index format: yyyy/mm/dd/name'


class Wondermark(WordPressScraper):
    url = 'http://wondermark.com/'
    stripUrl = url + '%s/'
    firstStripUrl = stripUrl % '001'
    prevSearch = '//a[@rel="prev"]'
    help = 'Index format: nnn (001-999), 1knn (1000-1099), cnnnn (1100-)'


class WorldOfMrToast(_BasicScraper):
    baseUrl = 'http://www.theimaginaryworld.com/'
    url = baseUrl + 'mrTcomicA.html'
    imageSearch = compile(tagre("img", "src", r'(comic[^"]+)'))
    # list the archive links since there is no prev/next navigation
    prevurls = (
        url,
        baseUrl + 'mrTcomicW02.html',
        baseUrl + 'mrTcomicW01.html',
        baseUrl + 'mrGcomic03.html',
        baseUrl + 'mrGcomic02.html',
        baseUrl + 'mrGcomic01.html',
        baseUrl + 'mrTcomicT05.html',
        baseUrl + 'mrTcomicT04.html',
        baseUrl + 'mrTcomicT03.html',
        baseUrl + 'mrTcomicT02.html',
        baseUrl + 'mrTcomicT01.html',
        baseUrl + 'mrTcomicIW3.html',
        baseUrl + 'mrTcomicIW2.html',
        baseUrl + 'mrTcomicIW1.html',
    )
    firstStripUrl = prevurls[-1]
    multipleImagesPerStrip = True
    endOfLife = True

    def getPrevUrl(self, url, data):
        idx = self.prevurls.index(url)
        try:
            return self.prevurls[idx + 1]
        except IndexError:
            return None


class WormWorldSaga(_BasicScraper):
    url = 'http://www.wormworldsaga.com/'
    stripUrl = url + 'chapters/%s/index.php'
    firstStripUrl = stripUrl % 'chapter01/EN'
    imageSearch = (
        compile(tagre("img", "src", r'(images/CH\d+_\d+\.[^"]+)')),
        compile(tagre("img", "src", r'(panels/CH\d+_[^"]+)')),
    )
    latestChapter = 5
    multipleImagesPerStrip = True

    def starter(self):
        return '%schapters/chapter%02d/%s/index.php' % (
            self.url, self.latestChapter, self.lang.upper())

    def getPrevUrl(self, url, data):
        """Find previous URL."""
        if 'chapter04' in url:
            return url.replace('chapter04', 'chapter03')
        if 'chapter03' in url:
            return url.replace('chapter03', 'chapter02')
        if 'chapter02' in url:
            return url.replace('chapter02', 'chapter01')
        return None


class WormWorldSagaFrench(WormWorldSaga):
    lang = 'fr'


class WormWorldSagaGerman(WormWorldSaga):
    lang = 'de'


class WormWorldSagaSpanish(WormWorldSaga):
    lang = 'es'



# SPDX-License-Identifier: MIT
# Copyright (C) 2004-2008 Tristan Seligmann and Jonathan Jacobs
# Copyright (C) 2012-2014 Bastian Kleineidam
# Copyright (C) 2015-2022 Tobias Gruetzmacher
# Copyright (C) 2019-2020 Daniel Ring
from ..scraper import ParserScraper

# Comicgenesis has a lot of comics, but most of them are disallowed by
# robots.txt


class ComicGenesis(ParserScraper):
    multipleImagesPerStrip = True
    imageSearch = '//img[contains(@src, "/comics/")]'
    prevSearch = (
        '//a[img/@alt="Previous comic"]',
        '//a[text()="Previous comic"]',
    )
    help = 'Index format: yyyymmdd'

    def __init__(self, name, sub=None, last=None, baseUrl=None, lang=None):
        super().__init__('ComicGenesis/' + name)

        if sub:
            baseUrl = 'http://%s.comicgenesis.com/' % sub

        self.stripUrl = baseUrl + 'd/%s.html'
        if last:
            self.url = self.stripUrl % last
            self.endOfLife = True
        else:
            self.url = baseUrl

        if lang:
            self.lang = lang

    @classmethod
    def getmodules(cls):
        return (
            cls('AAAAA', 'aaaaa'),
            cls('AdventuresofKiltman', 'kiltman'),
            cls('AmorModerno', 'amormoderno'),
            cls('AnythingButRealLife', 'anythingbutreallife'),
            cls('Ardra', 'ardra'),
            cls('Artwork', 'artwork'),
            cls('BabeintheWoods', 'babeinthewoods'),
            cls('BackwaterPlanet', 'bobthespirit'),
            cls('BendyStrawVampires', 'bsvampires'),
            cls('BlindSight', 'blindsight'),
            cls('BreakingtheDoldrum', 'breakingthedoldrum'),
            cls('BrotherSwan', 'warlordofnoodles'),
            cls('Candi', baseUrl='http://candicomics.com/'),
            cls('CorporateLife', 'corporatelife'),
            cls('DarkWelkin', 'darkwelkin'),
            cls('DeepBlue', 'gjbivin', last='20131109'),
            cls('DemonEater', 'demoneater'),
            cls('Dissonance', 'dissonance'),
            cls('DoodleDiaries', 'doodlediaries'),
            cls('DormSweetDorm', 'dormsweetdorm'),
            cls('DoubleyouTeeEff', 'doubleyouteeeff'),
            cls('DragonsBane', 'jasonwhitewaterz'),
            cls('Dreamaniac', 'dreamaniaccomic'),
            cls('ElnifiChronicles', 'elnifichronicles'),
            cls('EvesApple', 'evesapple'),
            cls('FancyThat', 'fancythat'),
            cls('FantasyQwest', 'creatorauthorman'),
            cls('Fantazine', 'fantazin'),
            cls('Flounderville', 'flounderville'),
            cls('GEM', 'keltzy'),
            cls('Gonefor300days', 'g4300d'),
            cls('ImpendingDoom', 'impending'),
            cls('InANutshell', 'nutshellcomics'),
            cls('KernyMantisComics', 'kernymantis'),
            cls('KitsuneJewel', 'kitsunejewel'),
            cls('KittyCattyGames', 'kittycattygames'),
            cls('KiwiDayN', 'kiwidayn'),
            cls('KungFounded', 'kungfounded'),
            cls('LabBratz', 'labbratz'),
            cls('Laserwing', 'laserwing'),
            cls('LumiasKingdom', 'lumia'),
            cls('Majestic7', 'majestic7'),
            cls('MaximumWhimsy', 'maximumwhimsy'),
            cls('MenschUnsererZeitGerman', 'muz', lang='de', last='20090630'),
            cls('MenschUnsererZeit', 'rabe', last='20090630'),
            cls('MoonCrest24', 'mooncrest', last='20121117'),
            cls('Mushian', 'tentoumushi'),
            cls('NightwolfCentral', 'nightwolfcentral'),
            cls('NoneMoreComic', 'nonemore'),
            cls('NoTimeForLife', 'randyraven', last='20100510'),
            cls('OcculTango', 'occultango'),
            cls('ODCKS', 'odcks'),
            cls('OfDoom', 'ofdoom'),
            cls('OpportunityofaLifetime', 'carpathia'),
            cls('Orbz', 'orbz'),
            cls('OwMySanity', 'owmysanity'),
            cls('PhantomThesis', 'phantomthesis'),
            cls('ProfessorSaltinesAstrodynamicDirigible', 'drsaltine'),
            cls('PsychicDyslexiaInstitute', 'pdi'),
            cls('PublicidadeEnganosa', 'publicidadeenganosa'),
            cls('RandomAxeOfKindness', 'randomaxe'),
            cls('SalemUncommons', 'salemuncommons'),
            cls('SamandElisAdventures', 'sameliadv'),
            cls('SarahZero', 'plughead'),
            cls('SixByNineCollege', 'sixbyninecollege'),
            cls('SpoononHighandFireontheMountian', 'spoon'),
            cls('SynapticMisfires', 'synapticmisfires'),
            cls('TakingStock', 'mapaghimagsik'),
            cls('TemplarArizona', 'templaraz'),
            cls('TheAdventuresofKaniraBaxter', 'kanirabaxter'),
            cls('TheAdventuresofVindibuddSuperheroInTraining', 'vindibudd', last='20070720'),
            cls('TheEasyBreather', 'easybreather'),
            cls('TheMisadventuresofOkk', 'okk'),
            cls('ThePath', 'thepath', '20081226'),
            cls('TheTalesofKalduras', 'kalduras'),
            cls('Unconventional', 'unconventional'),
            cls('WarMageNC17', 'warmage'),
            cls('WebcomicTheWebcomicWebcomicWebcomicWebcomic', 'dannormnsanidey'),
            cls('WhatYouDontSee', 'phantomlady4'),
            cls('Wierdman', 'asa'),
        )



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: © 2004 Tristan Seligmann and Jonathan Jacobs
# SPDX-FileCopyrightText: © 2012 Bastian Kleineidam
# SPDX-FileCopyrightText: © 2015 Tobias Gruetzmacher
# SPDX-FileCopyrightText: © 2019 Daniel Ring
from .. import util
from .common import WordPressNavi, WordPressScraper


class PetiteSymphony(WordPressNavi):
    multipleImagesPerStrip = True
    help = 'Index format: named number'

    def __init__(self, name):
        super(PetiteSymphony, self).__init__('PetiteSymphony/' +
                                             name.capitalize())
        self.url = 'http://%s.petitesymphony.com/' % name
        self.stripUrl = self.url + 'comic/%s'

    @classmethod
    def getmodules(cls):
        return (
            cls('knuckleup'),
            cls('sangria'),
        )


class ComicsBreak(WordPressScraper):
    def __init__(self, name, archive=None, adult=False):
        super(ComicsBreak, self).__init__('ComicsBreak/' + name)
        self.url = 'http://%s.comicsbreak.com/' % name.lower()
        if archive:
            self.url = 'https://web.archive.org/web/{}/{}'.format(
                archive, self.url)
            self.endOfLife = True
        if adult:
            self.adult = adult

    def namer(self, image_url, page_url):
        if self.name == 'ComicsBreak/Djandora':
            # Fix inconsistent filenames
            filename = util.urlpathsplit(image_url)[-1]
            filename = filename.replace('2014-10-31-Page70', 'Page70')
            filename = filename.replace('a3p69eng', 'Page69')
            if '2015/08/a4p57eng' in image_url:
                filename = filename.replace('p57', 'p56')
            return filename
        else:
            return super().namer(image_url, page_url)

    @classmethod
    def getmodules(cls):
        return (
            cls('Djandora', archive='20170923062433'),
            cls('Generation17', adult=True),
        )



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: © 2019 Tobias Gruetzmacher
# SPDX-FileCopyrightText: © 2019 Daniel Ring
from .. import scraper, util


class KemonoCafe(scraper.ParserScraper):
    imageSearch = '//div[@id="comic"]//img'
    prevSearch = '//a[contains(@class, "comic-nav-previous")]'

    def __init__(self, name, sub, first, last=None, adult=False):
        super(KemonoCafe, self).__init__('KemonoCafe/' + name)

        self.url = 'https://%s.kemono.cafe/' % sub
        self.stripUrl = self.url + 'comic/%s/'
        self.firstStripUrl = self.stripUrl % first

        if last:
            self.url = self.stripUrl % last
            self.endOfLife = True

        if adult:
            self.adult = True

    def namer(self, image_url, page_url):
        # Strip date from filenames
        filename = util.urlpathsplit(image_url)[-1]
        if 'ultrarosa' not in page_url:
            if filename[4] == '-' and filename[7] == '-':
                filename = filename[10:]
            if filename[0] == '-' or filename[0] == '_':
                filename = filename[1:]
        # Fix duplicate filenames
        if 'paprika' in page_url and '69-2' in page_url:
            filename = filename.replace('69', '69-2')
        elif 'rascals' in page_url and '89-2' in page_url:
            filename = filename.replace('89', '90')
        elif 'rascals' in page_url and '133-2' in page_url:
            filename = filename.replace('133', '134')
        elif 'caughtinorbit' in page_url and '26gs' in filename:
            filename = filename.replace('026gs', '021')
        elif 'caughtinorbit' in page_url and '27gs' in filename:
            filename = filename.replace('027gs', '022')
        # Fix unordered filenames
        if 'addictivescience' in page_url:
            page = self.getPage(page_url)
            num = int(self.match(page, '//div[@id="comic-wrap"]/@class')[0].replace('comic-id-', ''))
            filename = '%04d_%s' % (num, filename)
        elif 'CaughtInOrbit' in filename:
            filename = filename.replace('CaughtInOrbit', 'CIO')
        return filename

    @classmethod
    def getmodules(cls):
        return (
            cls('AddictiveScience', 'addictivescience', 'page0001'),
            cls('Bethellium', 'bethellium', 'c01p00'),
            cls('CaribbeanBlue', 'cb', 'page000', last='page325'),
            cls('CaughtInOrbit', 'caughtinorbit', 'comic-cover'),
            cls('IMew', 'imew', 'imew00', last='imew50'),
            cls('Knighthood', 'knighthood', 'kh0001'),
            cls('KnuckleUp', 'knuckle-up', 'page001', adult=True),
            cls('LasLindas', 'laslindas', 'll0001', adult=True),
            cls('Paprika', 'paprika', 'page000'),
            cls('PracticeMakesPerfect', 'pmp', 'title-001'),
            cls('Rascals', 'rascals', 'rascals-pg-0', adult=True),
            cls('TheEyeOfRamalach', 'theeye', 'theeye-page01'),
            cls('TinaOfTheSouth', 'tots', 'tos-01-01'),
            cls('UltraRosa', 'ultrarosa', 'pg001'),
        )



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: © 2020 Tobias Gruetzmacher
from lxml import etree


NS = {
    'd': 'https://dosage.rocks/xpath',
    're': 'http://exslt.org/regular-expressions',
}


def find_by_class(context, cls):
    attributes = context.context_node.attrib
    if 'class' in attributes:
        return cls in attributes['class'].split(' ')
    return False


def find_by_class_start(context, cls):
    """Matches elements where any class name starts with the specified string.
    This is a "workaround" for frameworks which append random stuff to class
    names when delivering static files.
    """
    attributes = context.context_node.attrib
    if 'class' in attributes:
        return any(clsname.startswith(cls) for clsname in attributes['class'].split(' '))
    return False


dosagens = etree.FunctionNamespace(NS['d'])
dosagens['class'] = find_by_class
dosagens['class_start'] = find_by_class_start



# SPDX-License-Identifier: MIT
# Copyright (C) 2019-2022 Tobias Gruetzmacher
# Copyright (C) 2019-2020 Daniel Ring
from ..scraper import ParserScraper


class NamirDeiter(ParserScraper):
    imageSearch = '//img[contains(@src, "comics/")]'
    prevSearch = ('//a[@rel="prev"]',
                  '//a[./img[contains(@src, "previous")]]',
                  '//a[contains(text(), "Previous")]')

    def __init__(self, name, baseUrl, first=None, last=None):
        if name == 'NamirDeiter':
            super(NamirDeiter, self).__init__(name)
        else:
            super(NamirDeiter, self).__init__('NamirDeiter/' + name)

        self.url = 'https://' + baseUrl + '/'
        self.stripUrl = self.url + 'comics/index.php?date=%s'

        if first:
            self.firstStripUrl = self.stripUrl % first
        else:
            self.firstStripUrl = self.url + 'comics/'

        if last:
            self.url = self.stripUrl % last
            self.endOfLife = True

    def link_modifier(self, fromurl, tourl):
        # Links are often absolute and keep jumping between http and https
        return tourl.replace('http:', 'https:').replace('/www.', '/')

    @classmethod
    def getmodules(cls):
        return (
            cls('ApartmentForTwo', 'apartmentfor2.com'),
            cls('NamirDeiter', 'namirdeiter.com', last='20150410'),
            cls('NicoleAndDerek', 'nicoleandderek.com'),
            cls('OneHundredPercentCat', 'ndunlimited.com/100cat', last='20121001'),
            cls('SpareParts', 'sparepartscomics.com', first='20031022', last='20080331'),
            cls('TheNDU', 'thendu.com'),
            cls('WonderKittens', 'wonderkittens.com'),
            cls('YouSayItFirst', 'yousayitfirst.com', first='20040220', last='20130125'),
        )


class UnlikeMinerva(ParserScraper):
    name = 'NamirDeiter/UnlikeMinerva'
    baseUrl = 'https://unlikeminerva.com/archive/index.php'
    stripUrl = baseUrl + '?week=%s'
    url = stripUrl % '127'
    firstStripUrl = stripUrl % '26'
    imageSearch = '//img[contains(@src, "archive/")]'
    prevSearch = '//a[./img[contains(@src, "previous")]]'
    multipleImagesPerStrip = True
    endOfLife = True



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: © 2025 Tobias Gruetzmacher

import logging
from typing import TYPE_CHECKING

MOREINFO = 15
TRACE = 5

if TYPE_CHECKING:
    parentlogger = logging.Logger
else:
    parentlogger = logging.getLoggerClass()

class ExtLogger(parentlogger):
    def moreinfo(self, msg, *args, **kwargs):
        if self.isEnabledFor(MOREINFO):
            self._log(MOREINFO, msg, args, **kwargs)

    def trace(self, msg, *args, **kwargs):
        if self.isEnabledFor(TRACE):
            self._log(TRACE, msg, args, **kwargs)


logging.addLevelName(MOREINFO, "MOREINFO")
logging.addLevelName(TRACE, "TRACE")
logging.setLoggerClass(ExtLogger)



# SPDX-License-Identifier: MIT
# Copyright (C) 2004-2008 Tristan Seligmann and Jonathan Jacobs
# Copyright (C) 2012-2014 Bastian Kleineidam
# Copyright (C) 2015-2022 Tobias Gruetzmacher
from re import compile, escape

from ..helpers import bounceStarter, joinPathPartsNamer
from ..scraper import BasicScraper, ParserScraper
from ..util import tagre
from .common import WordPressNavi


class ZapComic(ParserScraper):
    url = 'http://www.zapcomic.com/'
    imageSearch = '//img[d:class("comic-item")]'
    prevSearch = '//a[d:class("previous-comic-link")]'
    endOfLife = True


class Zapiro(ParserScraper):
    url = 'http://mg.co.za/zapiro/'
    starter = bounceStarter
    imageSearch = '//div[@id="cartoon"]/img'
    prevSearch = '//a[d:class("left")]'
    nextSearch = '//a[d:class("right")]'
    namer = joinPathPartsNamer(pageparts=(-1,))


class ZenPencils(WordPressNavi):
    url = 'https://web.archive.org/web/20200723091741/https://zenpencils.com/'
    multipleImagesPerStrip = True
    firstStripUrl = url + 'comic/1-ralph-waldo-emerson-make-them-cry/'
    starter = bounceStarter
    prevSearch = '//a[d:class("navi-prev")]'
    nextSearch = '//a[d:class("navi-next")]'
    endOfLife = True


class ZombieHunters(BasicScraper):
    url = 'http://www.thezombiehunters.com/'
    stripUrl = url + '?strip_id=%s'
    firstStripUrl = stripUrl % '1'
    imageSearch = compile(tagre("img", "src", r'(/istrip_files/strips/[^"]+)'))
    prevSearch = compile(tagre("a", "href", r'(\?strip_id=\d+)') + tagre("img", "id", "prevcomic"))
    help = 'Index format: n(unpadded)'


class Zwarwald(BasicScraper):
    url = "http://www.zwarwald.de/"
    rurl = escape(url)
    stripUrl = url + 'index.php/page/%s/'
    # anything before page 495 seems to be flash
    firstStripUrl = stripUrl % '495'
    lang = 'de'
    imageSearch = (
        compile(tagre("img", "src", r'(%simages/\d+/\d+/[^"]+)' % rurl)),
        compile(tagre("img", "src", r'(http://wp1163540\.wp190\.webpack\.hosteurope\.de/wordpress/images/\d+/\d+/[^"]+)')),
    )
    prevSearch = compile(tagre("a", "href", r'(%sindex\.php/page/\d+/)' % rurl) +
                         tagre("img", "src",
                               r'http://zwarwald\.de/images/prev\.jpg',
                               quote="'"))
    namer = joinPathPartsNamer(imageparts=range(-3, 0))
    help = 'Index format: number'

    def shouldSkipUrl(self, url, data):
        """Some pages have flash content."""
        return url in (
            self.stripUrl % "112",
            self.stripUrl % "222",
            self.stripUrl % "223",
            self.stripUrl % "246",
            self.stripUrl % "368",
            self.stripUrl % '495',
        )
